﻿using System;
using System.ComponentModel;
using System.Globalization;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;
using JetBrains.Annotations;
using Newtonsoft.Json;

namespace ECrash.Core.Models;

[JsonConverter(typeof(DateJsonConverter))]
[TypeConverter(typeof(DateTypeConverter))]
public struct Date : IEquatable<Date>, IXmlSerializable, IComparable<Date>
{
	private int _year;
	private int _month;
	private int _day;

	public Date(int year, int month, int day)
	{
		if (year < 1 || year > 9999) throw new ArgumentOutOfRangeException(nameof(year), "year must be between 1 and 9999");
		if (month < 1 || month > 12) throw new ArgumentOutOfRangeException(nameof(month), "month must be between 1 and 12");
		if (day < 1 || day > 31) throw new ArgumentOutOfRangeException(nameof(day), "day must be between 1 and 31");
		_year = year;
		_month = month;
		_day = day;
	}

	public Date(DateTime dateTime)
	{
		_year = dateTime.Year;
		_month = dateTime.Month;
		_day = dateTime.Day;
	}

	public Date(string date)
	{
		var parts = date.Split('-');
		_year = int.Parse(parts[0]);
		_month = int.Parse(parts[1]);
		_day = int.Parse(parts[2]);
	}

	public int Year
	{
		get { return _year; }
		[UsedImplicitly] private set { _year = value; }
	}

	public int Month
	{
		get { return _month; }
		[UsedImplicitly] private set { _month = value; }
	}

	public int Day
	{
		get { return _day; }
		[UsedImplicitly] private set { _day = value; }
	}
		
	public static explicit operator Date(DateTime dateTime)
	{
		return new Date(dateTime);
	}

	public static explicit operator DateTime(Date date)
	{
		return new DateTime(date.Year, date.Month, date.Day, 0, 0, 0, DateTimeKind.Unspecified);
	}

	public static explicit operator Date?(DateTime? date)
	{
		return date != null ? (Date?)(Date)date.Value : null;
	}

	public static explicit operator DateTime?(Date? date)
	{
		return date != null ? (DateTime?)(DateTime)date.Value : null;
	}

	public static explicit operator string(Date? date)
	{
		var dt = (DateTime?)date;
		return dt?.ToString("yyyy-MM-dd");
	}

	public int CompareTo(Date other)
	{
		int result = _year.CompareTo(other._year);
		if (result == 0) result = _month.CompareTo(other._month);
		if (result == 0) result = _day.CompareTo(other._day);
		return result;
	}

	public override string ToString()
	{
		return ((DateTime)this).ToShortDateString();
	}

	public string ToIsoString()
	{
		return $"{_year:0000}-{_month:00}-{_day:00}";
	}

	public bool Equals(Date other)
	{
		return _year == other._year && _month == other._month && _day == other._day;
	}

	public override bool Equals(object obj)
	{
		if (ReferenceEquals(null, obj)) return false;
		return obj is Date date && Equals(date);
	}

	// ReSharper disable NonReadonlyFieldInGetHashCode
	public override int GetHashCode()
	{
		unchecked
		{
			var hashCode = _year;
			hashCode = (hashCode * 397) ^ _month;
			hashCode = (hashCode * 397) ^ _day;
			return hashCode;
		}
	}
	// ReSharper restore NonReadonlyFieldInGetHashCode

	public static bool operator ==(Date left, Date right)
	{
		return left.Equals(right);
	}

	public static bool operator !=(Date left, Date right)
	{
		return !left.Equals(right);
	}

	XmlSchema IXmlSerializable.GetSchema()
	{
		return null;
	}

	void IXmlSerializable.ReadXml(XmlReader reader)
	{
		string text = reader.Value;
		var value = DateTime.ParseExact(text, "yyyy-MM-dd", CultureInfo.InvariantCulture);
		_year = value.Year;
		_month = value.Month;
		_day = value.Day;
		reader.Read();
	}

	void IXmlSerializable.WriteXml(XmlWriter writer)
	{
		writer.WriteString(((DateTime)this).ToString("yyyy-MM-dd"));
	}
}