﻿using System;
using System.Collections.Generic;
using JetBrains.Annotations;
using Newtonsoft.Json;

namespace ECrash.Core.Models.Document;

public abstract class AttachmentBase<TRoot> : DocumentChild<TRoot> where TRoot : DocumentRoot<TRoot>
{
	private byte[] _data;

	protected AttachmentBase(Hash hash, int length)
	{
		Hash = hash;
		Length = length;
	}

	protected AttachmentBase([NotNull] byte[] data)
	{
		Hash = Hash.ComputeHash(data);
		Length = data.Length;
		_data = data;
	}

	public override IEnumerable<IDocumentNode<TRoot>> EnumerateChildNodes()
	{
		yield break;
	}

	public abstract string AttachmentId { get; }
	public abstract Filename Filename { get; }
	public abstract string Type { get; set; }
	public abstract string Description { get; set; }
	public Hash Hash { get; }

	public int Length { get; }

	[JsonIgnore]
	public byte[] Data
	{
		get => _data;
		set
		{
			if (Equals(value, _data)) return;
			if (value == null)
			{
				throw new ArgumentNullException(nameof(value), "Data property cannot be explicitly set to null once it has been set to non-null.");
			}
			var computedHash = Hash.ComputeHash(value);
			if (value.Length != Length || computedHash != Hash)
			{
				throw new AttachmentCorruptedException(AttachmentId, Filename, Length, Hash, value.Length, computedHash);
			}
			_data = value;
			NotifyOfPropertyChange();
		}
	}

	public override void NotifyOfDocumentPropertyChange(DocumentPropertyChangedEventArgs args)
	{
		if (args.Source == this && args.PropertyName == "Data") return;
		base.NotifyOfDocumentPropertyChange(args);
	}

	public AttachmentKey GetKey() => new AttachmentKey(Root.Id, Filename, Hash);

	public static implicit operator AttachmentKey(AttachmentBase<TRoot> attachment) => attachment.GetKey();
}