﻿using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using ECrash.Core.Models.ReferenceData;

namespace ECrash.Core.Models.Document.BusinessRules;

public abstract class MutualExclusionBusinessRuleBase<TDocument> : DocumentBusinessRuleBase<TDocument> where TDocument : DocumentRoot<TDocument>
{
	protected MutualExclusionBusinessRuleBase(IReferenceDataCache referenceDataCache) : base(referenceDataCache)
	{
	}

	protected override void OnInitializeCore(TDocument document, ReferenceDataFilter referenceData)
	{
		foreach (var node in document.EnumerateNodes())
		{
			// If the node is a list of reference data values,
			// remove any non-exclusive items if any exclusive items are present
			if (node is IEnumerable<ReferenceDataBase> && node is IList list)
			{
				RemoveNonExclusiveItemsIfAnExclusiveItemIsPresent(list);
			}
		}
	}

	protected override void OnCollectionChangeCore(TDocument document, ReferenceDataFilter referenceData, DocumentCollectionChangedEventArgs change)
	{
		switch (change.Action)
		{
			case NotifyCollectionChangedAction.Add:
			case NotifyCollectionChangedAction.Replace:
				OnAddOrReplace(change);
				break;
			case NotifyCollectionChangedAction.Reset:
				OnReset(change);
				break;
		}
	}

	protected override void OnNodeAttached(TDocument document, IDocumentNode<TDocument> node, ReferenceDataFilter referenceData)
	{
		if (node is IEnumerable<ReferenceDataBase> && node is IList list)
		{
			RemoveNonExclusiveItemsIfAnExclusiveItemIsPresent(list);
		}
	}

	private void OnAddOrReplace(DocumentCollectionChangedEventArgs change)
	{
		var source = (IList)change.Source;
		var mutuallyExclusiveItem = change.NewItems.OfType<ReferenceDataBase>().FirstOrDefault(referenceData => referenceData.IsMutuallyExclusive);
		if (mutuallyExclusiveItem != null)
		{
			// A mutually exclusive item was added, so we remove everything but the mutually exclusive item
			for (int i = source.Count - 1; i >= 0; --i)
			{
				if (!ReferenceEquals(source[i], mutuallyExclusiveItem))
				{
					source.RemoveAt(i);
				}
			}
		}
		else
		{
			// A non-mutually exclusive item was added, so we remove all mutually exclusive items
			for (int i = source.Count - 1; i >= 0; --i)
			{
				var item = source[i] as ReferenceDataBase;
				if (item != null && item.IsMutuallyExclusive)
				{
					source.RemoveAt(i);
				}
			}
		}
	}

	private static void OnReset(DocumentCollectionChangedEventArgs change)
	{
		var source = (IList)change.Source;
		RemoveNonExclusiveItemsIfAnExclusiveItemIsPresent(source);
	}

	private static void RemoveNonExclusiveItemsIfAnExclusiveItemIsPresent(IList source)
	{
		var mutuallyExclusiveItem = source.OfType<ReferenceDataBase>().FirstOrDefault(referenceData => referenceData.IsMutuallyExclusive);
		if (mutuallyExclusiveItem == null) return;

		// If a mutually exclusive item was found, remove all items except the first mutually exclusive item
		for (int i = source.Count - 1; i >= 0; --i)
		{
			if (!ReferenceEquals(source[i], mutuallyExclusiveItem))
			{
				source.RemoveAt(i);
			}
		}
	}
}