﻿using System.Collections.Generic;
using System.Runtime.CompilerServices;
using JetBrains.Annotations;
using Newtonsoft.Json;

namespace ECrash.Core.Models.Document;

[JsonObject(MemberSerialization.OptOut)]
public abstract class DocumentNode<TRoot> : ExtendedPropertyChangedBase, IDocumentNode<TRoot>
	where TRoot : DocumentRoot<TRoot>
{
	[JsonIgnore]
	public abstract TRoot Root { get; protected set; }

	[JsonIgnore]
	public abstract IDocumentNode<TRoot> Parent { get; protected set; }

	public void UpdateLinks(IDocumentNode<TRoot> parent)
	{
		BeginUpdateLinks(parent);
		EndUpdateLinks(parent);
	}

	public virtual void BeginUpdateLinks(IDocumentNode<TRoot> parent)
	{
		Root?.NotifyOfNodeDetachment(this);

		Root = parent?.Root;
		Parent = parent;
		foreach (var child in EnumerateChildNodes())
		{
			child.BeginUpdateLinks(this);
		}
	}

	public virtual void EndUpdateLinks(IDocumentNode<TRoot> parent)
	{
		Root?.NotifyOfNodeAttachment(this);
		foreach (var child in EnumerateChildNodes())
		{
			child.EndUpdateLinks(this);
		}
	}

	public abstract IEnumerable<IDocumentNode<TRoot>> EnumerateChildNodes();

	public override void NotifyOfPropertyChange([CallerMemberName] string propertyName = null)
	{
		base.NotifyOfPropertyChange(propertyName);
		NotifyOfDocumentPropertyChange(new DocumentPropertyChangedEventArgs(this, propertyName));
	}

	public void NotifyOfNonDocumentPropertyChange([CallerMemberName] string propertyName = null)
	{
		base.NotifyOfPropertyChange(propertyName);
	}

	public abstract void NotifyOfDocumentPropertyChange(DocumentPropertyChangedEventArgs args);

	[NotifyPropertyChangedInvocator]
	public override bool Set<T>(ref T oldValue, T newValue, [CallerMemberName] string propertyName = null)
	{
		var oldChild = oldValue as DocumentChild<TRoot>;
		var newChild = newValue as DocumentChild<TRoot>;
		oldChild?.BeginUpdateLinks(null);
		newChild?.BeginUpdateLinks(this);
		var changed = base.Set(ref oldValue, newValue, propertyName);
		newChild?.EndUpdateLinks(this);
		oldChild?.EndUpdateLinks(null);

		return changed;
	}
}