﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading;
using Caliburn.Micro;
using ECrash.Core.Models.ReferenceData;
using JetBrains.Annotations;
using Newtonsoft.Json;

namespace ECrash.Core.Models.Document;

public abstract class DocumentRoot<TRoot> : DocumentNode<TRoot>
	where TRoot : DocumentRoot<TRoot>
{
	private static readonly IBusinessRule<TRoot>[] NoBusinessRules = Array.Empty<IBusinessRule<TRoot>>();
	private IList<IBusinessRule<TRoot>> _businessRules = NoBusinessRules;
	private int _suspensionCount;
	private ReferenceDataEpoch _referenceDataEpoch;

	public event EventHandler<DocumentPropertyChangedEventArgs> DocumentPropertyChanged;
	public event EventHandler<DocumentCollectionChangedEventArgs> DocumentCollectionChanged;

	protected DocumentRoot(Guid id)
	{
		Id = id;
		_referenceDataEpoch = new ReferenceDataEpoch(DateTime.Now, Assembly.GetExecutingAssembly().GetName().Version);
	}

	public Guid Id { get; }

	[JsonIgnore]
	public override TRoot Root
	{
		get => (TRoot)this;
		protected set => throw new NotSupportedException();
	}

	[JsonIgnore]
	public override IDocumentNode<TRoot> Parent
	{
		get => null;
		protected set => throw new NotSupportedException();
	}

	[JsonIgnore]
	public ReferenceDataEpoch ReferenceDataEpoch
	{
		get => _referenceDataEpoch;
		protected set => Set(ref _referenceDataEpoch, value);
	}

	[JsonIgnore]
	public virtual int Revision { get; set; }

	[JsonIgnore]
	public abstract int LatestRevision { get; set; }

	public override void BeginUpdateLinks(IDocumentNode<TRoot> parent)
	{
		throw new NotSupportedException();
	}

	public override void EndUpdateLinks(IDocumentNode<TRoot> parent)
	{
		throw new NotSupportedException();
	}

	public void Reinitialize()
	{
		_businessRules.Apply(rule => rule.OnInitialize((TRoot)this));
	}

	public virtual IEnumerable<AttachmentBase<TRoot>> EnumerateAttachments(Predicate<AttachmentBase<TRoot>> includePredicate = null)
	{
		yield break;
	}

	public IDisposable SuspendNotifications()
	{
		Interlocked.Increment(ref _suspensionCount);
		IsNotifying = false;
		return new DisposableAction(() =>
		{
			if (Interlocked.Decrement(ref _suspensionCount) == 0)
			{
				IsNotifying = true;
			}
		});
	}

	public void AttachBusinessRules([NotNull] IList<IBusinessRule<TRoot>> businessRules)
	{
		_businessRules = businessRules;
		using (SuspendNotifications())
		{
			Reinitialize();
		}
	}

	public void ApplyBusinessRulesOnce([NotNull] IList<IBusinessRule<TRoot>> businessRules)
	{
		using (SuspendNotifications())
		{
			businessRules.Apply(x => x.OnInitialize((TRoot)this));
		}
	}

	public IList<IBusinessRule<TRoot>> DetachBusinessRules()
	{
		var businessRules = _businessRules;
		_businessRules = NoBusinessRules;
		return businessRules;
	}

	/// <summary>
	/// Enumerates all IDocumentNode&lt;TRoot&gt; instances in the document.
	/// </summary>
	/// <returns></returns>
	public IEnumerable<IDocumentNode<TRoot>> EnumerateNodes()
	{
		var stack = new Stack<IDocumentNode<TRoot>>();
		var hashSet = new HashSet<IDocumentNode<TRoot>>();
		stack.Push(this);
		while (stack.Count > 0)
		{
			// Pop the node off the top of the stack
			var model = stack.Pop();
			// If we've seen it before, skip it
			if (!hashSet.Add(model)) continue;
			// Return it from the enumeration
			yield return model;
			// Find child nodes referenced via properties
			var properties = model.GetType().GetProperties().Where(x => typeof(IDocumentNode<TRoot>).IsAssignableFrom(x.PropertyType) && x.GetIndexParameters().Length == 0);
			foreach (var property in properties)
			{
				if (property.GetValue(model) is IDocumentNode<TRoot> node)
				{
					stack.Push(node);
				}
			}
			// Find child nodes referenced in collections
			if (model is IEnumerable<IDocumentNode<TRoot>> collection)
			{
				foreach (var node in collection)
				{
					stack.Push(node);
				}
			}
		}
	}

	public override void NotifyOfDocumentPropertyChange(DocumentPropertyChangedEventArgs args)
	{
		if (IsNotifying)
		{
			DocumentPropertyChanged?.Invoke(this, args);
			_businessRules.Apply(businessRule => businessRule.OnPropertyChange((TRoot)this, args));
		}
	}

	protected internal virtual void NotifyOfDocumentCollectionChange(DocumentCollectionChangedEventArgs args)
	{
		if (IsNotifying)
		{
			DocumentCollectionChanged?.Invoke(this, args);
			_businessRules.Apply(businessRule => businessRule.OnCollectionChange((TRoot)this, args));
		}
	}

	protected internal virtual void NotifyOfNodeAttachment(IDocumentNode<TRoot> node)
	{
		if (IsNotifying)
		{
			_businessRules.Apply(businessRule => businessRule.OnNodeAttached((TRoot)this, node));
		}
	}

	protected internal virtual void NotifyOfNodeDetachment(IDocumentNode<TRoot> node)
	{
		if (IsNotifying)
		{
			_businessRules.Apply(businessRule => businessRule.OnNodeDetached((TRoot)this, node));
		}
	}

	public void SynchronizeReferenceData(ReferenceDataList referenceData)
	{
		var exceptions = new List<Exception>();
		var businessRules = DetachBusinessRules();
		try
		{
			foreach (var node in EnumerateNodes())
			{
				if (node is IList list)
				{
					for (int i = 0; i < list.Count; ++i)
					{
						if (list[i] is ReferenceDataBase value)
						{
							string typeName = value.GetType().Name;
							var updatedValue = referenceData.GetById(typeName, value.Id);
							if (updatedValue == null)
							{
								exceptions.Add(new ReferenceDataMissingException(value));
								continue;
							}
							list[i] = updatedValue;
						}
					}
				}
				else
				{
					var type = node.GetType();
					var properties = type.GetProperties(BindingFlags.Public | BindingFlags.Instance);
					foreach (var property in properties.Where(x => x.CanRead && x.CanWrite && typeof(ReferenceDataBase).IsAssignableFrom(x.PropertyType)))
					{
						var value = (ReferenceDataBase)property.GetValue(node);
						if (value != null)
						{
							var valueType = value.GetType();
							string typeName = valueType.Name;
							var databaseValue = referenceData.GetById(typeName, value.Id);
							if (databaseValue == null)
							{
								exceptions.Add(new ReferenceDataMissingException(value));
								continue;
							}
							property.SetValue(node, databaseValue);
						}
					}
				}
			}
		}
		catch (Exception ex)
		{
			exceptions.Add(ex);
		}
		finally
		{
			AttachBusinessRules(businessRules);
		}
		if (exceptions.Count > 0)
		{
			throw new AggregateException("One or more reference data values specified in the report were not found in the database.", exceptions);
		}
	}
}