﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Threading;
using Caliburn.Micro;

namespace ECrash.Core.Models.Document;

public class DocumentValueCollection<TRoot, TValue> : BindableCollection<TValue>, IDocumentNode<TRoot>
	where TRoot : DocumentRoot<TRoot>
{
	private int _suspensionCount;

	[Newtonsoft.Json.JsonIgnore]
	public virtual TRoot Root { get; protected set; }

	[Newtonsoft.Json.JsonIgnore]
	public virtual IDocumentNode<TRoot> Parent { get; protected set; }

	public void UpdateLinks(IDocumentNode<TRoot> parent)
	{
		BeginUpdateLinks(parent);
		EndUpdateLinks(parent);
	}

	public virtual void BeginUpdateLinks(IDocumentNode<TRoot> parent)
	{
		Root = parent.Root;
		Parent = parent;
		foreach (var child in EnumerateChildNodes())
		{
			child.BeginUpdateLinks(this);
		}
	}

	public virtual void EndUpdateLinks(IDocumentNode<TRoot> parent)
	{
		Root?.NotifyOfNodeAttachment(this);
		foreach (var child in EnumerateChildNodes())
		{
			child.EndUpdateLinks(this);
		}
	}

	public virtual IEnumerable<IDocumentNode<TRoot>> EnumerateChildNodes()
	{
		yield break;
	}

	public override void AddRange(IEnumerable<TValue> items)
	{
		// Reimplemented this to use NotifyCollectionChangedAction.Add instead of Reset
		OnUIThread(() =>
		{
			int startIndex = Count;
			var newItemsList = items.ToList();
			using (SuspendNotifications())
			{
				foreach (var item in newItemsList)
				{
					InsertItemBase(Count, item);
				}
			}
			OnPropertyChanged(new PropertyChangedEventArgs("Count"));
			OnPropertyChanged(new PropertyChangedEventArgs("Item[]"));
			OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, newItemsList, startIndex));
		});
	}

	public IDisposable SuspendNotifications()
	{
		Interlocked.Increment(ref _suspensionCount);
		IsNotifying = false;
		return new DisposableAction(() =>
		{
			if (Interlocked.Decrement(ref _suspensionCount) == 0)
			{
				IsNotifying = true;
			}
		});
	}

	protected override void OnPropertyChanged(System.ComponentModel.PropertyChangedEventArgs e)
	{
		base.OnPropertyChanged(e);
		if (IsNotifying)
		{
			Root?.NotifyOfDocumentPropertyChange(new DocumentPropertyChangedEventArgs(this, e.PropertyName));
		}
	}

	protected override void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
	{
		base.OnCollectionChanged(e);
		if (IsNotifying)
		{
			Root?.NotifyOfDocumentCollectionChange(new DocumentCollectionChangedEventArgs(this, e));
		}
	}
}