﻿using System;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.Threading.Tasks;
using JetBrains.Annotations;
using Newtonsoft.Json;

namespace ECrash.Core.Models.Document;

[JsonConverter(typeof(HashJsonConverter))]
public struct Hash : IEquatable<Hash>
{
	private readonly byte[] _hash;

	public Hash([NotNull] byte[] hash)
	{
		_hash = hash ?? throw new ArgumentNullException(nameof(hash));
	}

	public Hash([NotNull] string hash)
	{
		if (hash == null) throw new ArgumentNullException(nameof(hash));
		_hash = hash.FromHexString();
	}

	public static Hash ComputeHash([NotNull] byte[] data)
	{
		if (data == null) throw new ArgumentNullException(nameof(data));
		using (var sha1 = SHA1.Create())
		{
			return new Hash(sha1.ComputeHash(data));
		}
	}

	public static Hash ComputeHash([NotNull] Stream stream)
	{
		if (stream == null) throw new ArgumentNullException(nameof(stream));
		using (var sha1 = SHA1.Create())
		{
			return new Hash(sha1.ComputeHash(stream));
		}
	}

	public static async Task<Hash> ComputeHashAsync([NotNull] Stream stream)
	{
		if (stream == null) throw new ArgumentNullException(nameof(stream));
		using (var sha1 = SHA1.Create())
		{
			var buffer = new byte[8192];
			int bytesRead;

			while ((bytesRead = await stream.ReadAsync(buffer, 0, buffer.Length).ConfigureAwait(false)) != 0)
			{
				sha1.TransformBlock(buffer, 0, bytesRead, buffer, 0);
			}
			sha1.TransformFinalBlock(buffer, 0, bytesRead);
			return new Hash(sha1.Hash);
		}
	}

	public override string ToString()
	{
		return _hash.ToHexString();
	}

	public static explicit operator string(Hash hash) => hash.ToString();
	public static explicit operator Hash(string hash) => new Hash(hash);

	public bool Equals(Hash other)
	{
		if (_hash.Length != other._hash.Length) return false;
		for (int i = 0; i < _hash.Length; ++i)
		{
			if (_hash[i] != other._hash[i]) return false;
		}
		return true;
	}

	public override bool Equals(object obj)
	{
		return obj is Hash other && Equals(other);
	}

	public override int GetHashCode()
	{
		unchecked
		{
			return _hash.Aggregate(0, (current, b) => (current * 397) ^ b);
		}
	}

	public static bool operator ==(Hash left, Hash right)
	{
		return left.Equals(right);
	}

	public static bool operator !=(Hash left, Hash right)
	{
		return !left.Equals(right);
	}
}