﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace ECrash.Core.Models.ReferenceData;

[DebuggerDisplay("[{Code}] {Name}")]
[JsonObject(MemberSerialization.OptIn)]
public abstract class ReferenceDataBase : IEquatable<ReferenceDataBase>, ICloneable
{
	protected ReferenceDataBase(int id, RowVersion version)
	{
		Id = id;
		Version = version;
	}

	[JsonProperty] public int Id { get; }
	[JsonProperty] public RowVersion Version { get; }
	[JsonProperty] public string Code { get; set; }
	[JsonProperty] public string Name { get; set; }
	[JsonProperty] public string Category { get; set; }
	[JsonProperty] public double Order { get; set; }
	[JsonProperty] public DateTime? EffectiveDate { get; set; }
	[JsonProperty] public DateTime? ExpirationDate { get; set; }
	[JsonProperty] public Version EffectiveVersion { get; set; }
	[JsonProperty] public Version ExpirationVersion { get; set; }
	[JsonProperty] public ISet<string> Tags { get; } = new HashSet<string>();
	[JsonProperty] public IDictionary<string, JToken> Fields { get; } = new Dictionary<string, JToken>();

	[JsonIgnore]
	public bool RequiresDescription
	{
		get => Tags.Contains(CommonTags.RequiresDescriptionTag);
		set => SetTagPresence(CommonTags.RequiresDescriptionTag, value);
	}

	[JsonIgnore]
	public bool IsMutuallyExclusive
	{
		get => Tags.Contains(CommonTags.MutuallyExclusiveTag);
		set => SetTagPresence(CommonTags.MutuallyExclusiveTag, value);
	}

	[JsonIgnore]
	public bool IsNotApplicable
	{
		get => Tags.Contains(CommonTags.NotApplicableTag);
		set => SetTagPresence(CommonTags.NotApplicableTag, value);
	}

	[JsonIgnore]
	public bool IsUnknown
	{
		get => Tags.Contains(CommonTags.UnknownTag);
		set => SetTagPresence(CommonTags.UnknownTag, value);
	}

	[JsonIgnore]
	public bool IsOther
	{
		get => Tags.Contains(CommonTags.OtherTag);
		set => SetTagPresence(CommonTags.OtherTag, value);
	}

	[JsonIgnore]
	public bool IsYes
	{
		get => Tags.Contains(CommonTags.YesTag);
		set => SetTagPresence(CommonTags.YesTag, value);
	}

	[JsonIgnore]
	public bool IsNo
	{
		get => Tags.Contains(CommonTags.NoTag);
		set => SetTagPresence(CommonTags.NoTag, value);
	}

	public bool IsInEffect(ReferenceDataEpoch epoch)
	{
		bool effectiveByDate = (EffectiveDate == null || epoch.Timestamp >= EffectiveDate.Value) 
		                       && (ExpirationDate == null || epoch.Timestamp < ExpirationDate.Value);
		bool effectiveByVersion = (EffectiveVersion == null || epoch.Version>= EffectiveVersion) 
		                          && (ExpirationVersion == null || epoch.Version < ExpirationVersion);
		return effectiveByDate && effectiveByVersion;
	}

	public ReferenceDataBase VersionedClone() => Clone(true);
	public ReferenceDataBase UnversionedClone() => Clone(false);

	protected ReferenceDataBase Clone(bool keepVersion)
	{
		var value = (ReferenceDataBase)Activator.CreateInstance(GetType(), Id, keepVersion ? Version : RowVersion.Empty);
		value.Code = Code;
		value.Name = Name;
		value.Category = Category;
		value.Order = Order;
		value.EffectiveDate = EffectiveDate;
		value.ExpirationDate = ExpirationDate;
		value.EffectiveVersion = EffectiveVersion;
		value.ExpirationVersion = ExpirationVersion;
		foreach (string tag in Tags)
		{
			value.Tags.Add(tag);
		}
		foreach (var field in Fields)
		{
			value.Fields.Add(field.Key, field.Value?.DeepClone());
		}
		return value;
	}

	protected void SetTagPresence(string tag, bool presence)
	{
		if (presence)
		{
			Tags.Add(tag);
		}
		else
		{
			Tags.Remove(tag);
		}
	}

	public T GetFieldObject<T>(string fieldName, T defaultValue = default)
	{
		return Fields.TryGetValue(fieldName, out var token) ? token.ToObject<T>() : defaultValue;
	}

	public T GetFieldValue<T>(string fieldName, T defaultValue = default)
	{
		return Fields.TryGetValue(fieldName, out var token) ? token.Value<T>() : defaultValue;
	}

	public IEnumerable<T> GetFieldValues<T>(string fieldName)
	{
		if (!Fields.TryGetValue(fieldName, out var token) || token is not JArray array) yield break;
		foreach (var item in array)
		{
			yield return item.ToObject<T>();
		}
	}

	public bool Equals(ReferenceDataBase other)
	{
		if (ReferenceEquals(null, other)) return false;
		if (ReferenceEquals(this, other)) return true;
		return Id == other.Id && Version == other.Version;
	}

	public override bool Equals(object obj)
	{
		if (ReferenceEquals(null, obj)) return false;
		if (ReferenceEquals(this, obj)) return true;
		if (obj.GetType() != GetType()) return false;
		return Equals((ReferenceDataBase)obj);
	}

	public bool IsEquivalentTo(ReferenceDataBase other, bool checkVersion = false)
	{
		if (ReferenceEquals(null, other)) return false;
		if (ReferenceEquals(this, other)) return true;
		bool equivalent = Id == other.Id
		                  && (!checkVersion || Version == other.Version)
		                  && string.Equals(Code, other.Code, StringComparison.Ordinal)
		                  && string.Equals(Name, other.Name, StringComparison.Ordinal)
		                  && string.Equals(Category, other.Category, StringComparison.Ordinal)
		                  // ReSharper disable once CompareOfFloatsByEqualityOperator
		                  && Order == other.Order
		                  && EffectiveDate == other.EffectiveDate
		                  && ExpirationDate == other.ExpirationDate
		                  && EffectiveVersion == other.EffectiveVersion
		                  && ExpirationVersion == other.ExpirationVersion
		                  && Tags.SetEquals(other.Tags)
		                  && Fields.Count == other.Fields.Count;
		if (!equivalent) return false;
		foreach (var field in Fields)
		{
			if (!other.Fields.TryGetValue(field.Key, out var otherValue)) return false;
			if (!field.Value.ToString(Formatting.None).Equals(otherValue.ToString(Formatting.None), StringComparison.Ordinal)) return false;
		}
		return true;
	}

	public override int GetHashCode()
	{
		unchecked
		{
			return (Id * 397) ^ Version.GetHashCode();
		}
	}

	object ICloneable.Clone()
	{
		return Clone(true);
	}

	public static bool operator ==(ReferenceDataBase left, ReferenceDataBase right)
	{
		return Equals(left, right);
	}

	public static bool operator !=(ReferenceDataBase left, ReferenceDataBase right)
	{
		return !Equals(left, right);
	}
}