﻿using System;
using System.Collections.Generic;
using System.Linq;
using Newtonsoft.Json;

namespace ECrash.Core.Models.ReferenceData;

[JsonConverter(typeof(ReferenceDataListJsonConverter))]
public class ReferenceDataList : ICloneable
{
	private readonly Dictionary<string, List<ReferenceDataBase>> _referenceData;
	private bool _isInitializing;

	public ReferenceDataList()
	{
		_referenceData = new Dictionary<string, List<ReferenceDataBase>>();
		foreach (var type in ReferenceDataMetadata.Types)
		{
			_referenceData[type.Name] = new List<ReferenceDataBase>();
		}
	}

	public ReferenceDataList(Dictionary<string, List<ReferenceDataBase>> referenceData)
	{
		_referenceData = referenceData;
	}

	public IDictionary<string, List<ReferenceDataBase>> ReferenceData => _referenceData;

	public RowVersion Version => _referenceData.SelectMany(x => x.Value).MaxOrDefault(x => x.Version, RowVersion.Empty);

	public void BeginInit()
	{
		_isInitializing = true;
	}

	public void EndInit()
	{
		_isInitializing = false;
	}

	public IList<T> GetAllDataOfType<T>() where T : ReferenceDataBase
	{
		return GetProcessedList<T>(items => items.OrderBy(refData => refData.Order));
	}

	public IList<ReferenceDataBase> GetAllDataOfType(Type type)
	{
		return GetProcessedList(type, items => items.OrderBy(refData => refData.Order));
	}

	public IList<T> GetDataInEffectAsOf<T>(ReferenceDataEpoch epoch) where T : ReferenceDataBase
	{
		return GetProcessedList<T>(items => items.Where(refData => refData.IsInEffect(epoch)).OrderBy(refData => refData.Order));
	}

	public T GetById<T>(int id) where T : ReferenceDataBase
	{
		var listForType = EnumerateValuesForType<T>();
		return listForType.FirstOrDefault(rd => rd.Id == id) as T;
	}

	public ReferenceDataBase GetById(string typeName, int id)
	{
		var listForType = GetListForType(typeName);
		return listForType.FirstOrDefault(rd => rd.Id == id);
	}

	public void Add(string typeName, ReferenceDataBase item)
	{
		if (!_isInitializing) item = item.UnversionedClone();
		var listForType = GetListForType(typeName);
		var existingValue = listForType.FirstOrDefault(x => x.Id == item.Id);
		if (existingValue != null)
		{
			if (existingValue.IsEquivalentTo(item))
			{
				// New item appears to be the same as the existing item, so don't add it
				return;
			}
			// New item conflicts with existing item, so throw an exception
			throw new ArgumentException($"The reference data list already contains an item of type {typeName} with ID {item.Id}");
		}
		listForType.Add(item);
	}

	public void Add<T>(T item) where T : ReferenceDataBase
	{
		string typeName = item.GetType().Name;
		Add(typeName, item);
	}

	public bool AddIfNotPresent(string typeName, ReferenceDataBase item)
	{
		var listForType = GetListForType(typeName);
		var existingValue = listForType.FirstOrDefault(x => x.Id == item.Id);
		if (existingValue != null) return false;
		listForType.Add(item);
		return true;
	}

	public bool AddIfNotPresent<T>(T item) where T : ReferenceDataBase
	{
		string typeName = item.GetType().Name;
		return AddIfNotPresent(typeName, item);
	}

	public void AddOrReplace(string typeName, ReferenceDataBase item)
	{
		item = item.UnversionedClone();
		var listForType = GetListForType(typeName);
		for (int i = 0; i < listForType.Count; ++i)
		{
			var existingItem = listForType[i];
			if (existingItem.Id == item.Id)
			{
				if (!item.IsEquivalentTo(existingItem)) listForType[i] = item;
				return;
			}
		}
		listForType.Add(item);
	}

	public void AddOrReplace<T>(T item) where T : ReferenceDataBase
	{
		string typeName = item.GetType().Name;
		AddOrReplace(typeName, item);
	}

	public void Update<T>(int id, Action<T> update) where T : ReferenceDataBase
	{
		var item = GetById<T>(id);
		Update(item, update);
	}

	public void Update<T>(T item, Action<T> update) where T : ReferenceDataBase
	{
		var copy = (T)item.UnversionedClone();
		update(copy);
		AddOrReplace(copy);
	}

	private IList<ReferenceDataBase> GetProcessedList(Type type, Func<IEnumerable<ReferenceDataBase>, IEnumerable<ReferenceDataBase>> processingFunction)
	{
		return processingFunction(EnumerateValuesForType(type)).ToList();
	}

	private IList<T> GetProcessedList<T>(Func<IEnumerable<T>, IEnumerable<T>> processingFunction) where T : ReferenceDataBase
	{
		return processingFunction(EnumerateValuesForType<T>().Cast<T>()).ToList();
	}

	private IEnumerable<ReferenceDataBase> EnumerateValuesForType(Type type)
	{
		return GetListForType(type.Name);
	}

	private IEnumerable<ReferenceDataBase> EnumerateValuesForType<T>() where T : ReferenceDataBase
	{
		return GetListForType(typeof(T).Name);
	}

	private IList<ReferenceDataBase> GetListForType(string typeName)
	{
		if (!_referenceData.TryGetValue(typeName, out var listForType))
		{
			listForType = new List<ReferenceDataBase>();
			_referenceData.Add(typeName, listForType);
		}
		return listForType;
	}

	public ReferenceDataList UnversionedClone() => Clone(false);
	public ReferenceDataList VersionedClone() => Clone(true);

	protected ReferenceDataList Clone(bool keepVersion)
	{
		var clone = new ReferenceDataList();
		clone.BeginInit();
		foreach (var type in _referenceData.Keys)
		{
			foreach (var value in _referenceData[type])
			{
				clone.Add(keepVersion ? value.VersionedClone() : value.UnversionedClone());
			}
		}
		clone.EndInit();
		return clone;
	}

	object ICloneable.Clone()
	{
		return UnversionedClone();
	}
}