﻿using System;
using System.ComponentModel;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;
using JetBrains.Annotations;
using Newtonsoft.Json;

namespace ECrash.Core.Models;

[JsonConverter(typeof(TimeJsonConverter))]
[TypeConverter(typeof(TimeTypeConverter))]
public struct Time : IXmlSerializable, IEquatable<Time>, IComparable<Time>
{
	public Time(int hour, int minute)
	{
		if (hour < 0 || hour > 23) throw new ArgumentOutOfRangeException(nameof(hour), "hour must be between 0 and 23");
		if (minute < 0 || minute > 59) throw new ArgumentOutOfRangeException(nameof(minute), "minute must be between 0 and 59");
		Hour = hour;
		Minute = minute;
	}

	public Time(DateTime dateTime)
	{
		Hour = dateTime.Hour;
		Minute = dateTime.Minute;
	}

	public Time(string time)
	{
		var parts = time.Split(':');
		Hour = int.Parse(parts[0]);
		Minute = int.Parse(parts[1]);
	}

	public int Hour { get; [UsedImplicitly] private set; }
	public int Minute { get; [UsedImplicitly] private set; }

	public Time AddMinutes(int minutes, out int carryDays)
	{
		if (minutes == 0)
		{
			carryDays = 0;
			return this;
		}

		int newMinute = Minute + minutes;
		int div = Math.DivRem(newMinute, 60, out int rem);
		if (rem < 0)
		{
			rem += 60;
			div -= 1;
		}
		newMinute = rem;
		int newHour = Hour + div;
		div = Math.DivRem(newHour, 24, out rem);
		if (rem < 0)
		{
			rem += 24;
			div -= 1;
		}
		newHour = rem;
		carryDays = div;
		return new Time(newHour, newMinute);
	}

	public static explicit operator Time(DateTime dateTime)
	{
		return new Time(dateTime);
	}

	public static explicit operator Time(TimeSpan timeSpan)
	{
		return new Time(timeSpan.Hours, timeSpan.Minutes);
	}

	public static explicit operator DateTime(Time time)
	{
		return new DateTime(2000, 1, 1, time.Hour, time.Minute, 0, DateTimeKind.Unspecified);
	}

	public static explicit operator Time?(DateTime? time)
	{
		return time != null ? (Time)time.Value : (Time?)null;
	}

	public static explicit operator DateTime?(Time? time)
	{
		return time != null ? (DateTime)time.Value : (DateTime?)null;
	}

	public static explicit operator TimeSpan(Time time)
	{
		return TimeSpan.FromMinutes(60 * time.Hour + time.Minute);
	}

	public static explicit operator TimeSpan?(Time? time)
	{
		return time != null ? (TimeSpan)time.Value : (TimeSpan?)null;
	}

	public static explicit operator Time?(TimeSpan? timeSpan)
	{
		return timeSpan != null ? (Time)timeSpan.Value : (Time?)null;
	}

	public static explicit operator string(Time? time)
	{
		var dt = (DateTime?)time;
		return dt?.ToString("HH:mm");
	}

	public int CompareTo(Time other)
	{
		int result = Hour.CompareTo(other.Hour);
		if (result == 0) result = Minute.CompareTo(other.Minute);
		return result;
	}

	public override string ToString()
	{
		return ((DateTime)this).ToShortTimeString();
	}

	public string ToIsoString()
	{
		return $"{Hour:00}:{Minute:00}:00";
	}

	public bool Equals(Time other)
	{
		return Hour == other.Hour && Minute == other.Minute;
	}

	public override bool Equals(object obj)
	{
		if (ReferenceEquals(null, obj)) return false;
		return obj is Time time && Equals(time);
	}

	// ReSharper disable NonReadonlyFieldInGetHashCode
	public override int GetHashCode()
	{
		unchecked
		{
			return (Hour * 397) ^ Minute;
		}
	}
	// ReSharper restore NonReadonlyFieldInGetHashCode

	XmlSchema IXmlSerializable.GetSchema()
	{
		return null;
	}

	void IXmlSerializable.ReadXml(XmlReader reader)
	{
		string text = reader.Value;
		var value = XmlConvert.ToDateTime(text, "HH:mm:ss");
		Hour = value.Hour;
		Minute = value.Minute;
		reader.Read();
	}

	void IXmlSerializable.WriteXml(XmlWriter writer)
	{
		writer.WriteString(ToIsoString());
	}
}