﻿using System;
using System.Collections.Generic;

namespace ECrash.Core.Models;

public class UtmGeographicBorder
{
	public List<UtmGeographicPolygon> Polygons { get; } = new();

	public bool IsCoordinateWithin(double latitude, double longitude)
	{
		int lastZone = 0;
		double x = 0;
		double y = 0;
		foreach (var polygon in Polygons)
		{
			if (polygon.Zone != lastZone)
			{
				lastZone = ConvertCoordinatesToUtm(latitude, longitude, out x, out y, polygon.Zone);
			}
			if (polygon.IsPointWithin(x, y)) return true;
		}

		return false;
	}

	/// <summary>
	/// Converts latitude and longitude coordinates to Unversal Transverse Mercator coordinates.
	/// <see cref="http://en.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system"/>
	/// </summary>
	/// <param name="latitude">The latitude coordinate.</param>
	/// <param name="longitude">The longitude coordinate.</param>
	/// <param name="easting">The UTM easting coordinate.</param>
	/// <param name="northing">The UTM northing coordinate.</param>
	private static int ConvertCoordinatesToUtm(double latitude, double longitude, out double easting, out double northing, int zone = 0)
	{
		bool isWesternHemisphere = longitude < 0;
		double latRad = latitude * Math.PI / 180;
		if (zone <= 0)
		{
			zone = isWesternHemisphere ? (int)((180 + longitude) / 6) + 1 : (int)(longitude / 6) + 31;
		}
		double long0 = 6 * zone - 183;

		const double a = 6378137;
		const double b = 6356752.314;
		const double n = (a - b) / (a + b);
		const double c0 = (15 * a * n * n / 16) * (1 - n + (3 * n * n / 4) * (1 - n));
		const double k0 = 0.9996;
		const double sin1Sec = Math.PI / (180 * 3600);

		double e = Math.Sqrt(1 - Math.Pow(b, 2) / Math.Pow(a, 2));
		double eprimesq = Math.Pow(e * a / b, 2);
		double nu = a / Math.Pow(1 - Math.Pow(e, 2) * Math.Pow(Math.Sin(latRad), 2), 1.0 / 2);
		double p = (longitude - long0) * 3600 / 10000;

		double a0 = a * (1 - n + (5 * n * n / 4.0) * (1 - n) + (81.0 * Math.Pow(n, 4.0) / 64) * (1 - n));
		double b0 = (3 * a * n / 2) * (1 - n - (7 * n * n / 8) * (1 - n) + (55 * Math.Pow(n, 4.0) / 64.0));
		double d0 = (35 * a * Math.Pow(n, 3.0) / 48.0) * (1 - n + 11 * n * n / 16);
		double e0 = (315 * a * Math.Pow(n, 4.0) / 51.0) * (1 - n);

		double s = a0 * latRad - b0 * Math.Sin(2 * latRad) + c0 * Math.Sin(4 * latRad) - d0 * Math.Sin(6 * latRad) + e0 * Math.Sin(8 * latRad);

		double k1 = s * k0;
		double k2 = nu * Math.Sin(latRad) * Math.Cos(latRad) * Math.Pow(sin1Sec, 2) * k0 * (100000000) / 2;
		double k3 = ((Math.Pow(sin1Sec, 4) * nu * Math.Sin(latRad) * Math.Pow(Math.Cos(latRad), 3)) / 24) *
		            (5 - Math.Pow(Math.Tan(latRad), 2) + 9 * eprimesq * Math.Pow(Math.Cos(latRad), 2) +
		             4 * Math.Pow(eprimesq, 2) * Math.Pow(Math.Cos(latRad), 4)) * k0 * (10000000000000000);
		double k4 = nu * Math.Cos(latRad) * sin1Sec * k0 * 10000;
		double k5 = Math.Pow((sin1Sec * Math.Cos(latRad)), 3) * (nu / 6) *
		            (1 - Math.Pow(Math.Tan(latRad), 2) + eprimesq * Math.Pow(Math.Cos(latRad), 2)) * k0 * (1000000000000);

		northing = k1 + k2 * Math.Pow(p, 2) + k3 * Math.Pow(p, 4);
		easting = 500000 + k4 * p + k5 * Math.Pow(p, 3);
		return zone;
	}
}