﻿using System;
using ECrash.Core.Models.Document;
using JetBrains.Annotations;
using Newtonsoft.Json;

namespace ECrashLA.Core.Models.CrashReports
{
	public class Attachment : AttachmentBase<CrashReport>, IEquatable<Attachment>, INeedsCollectionIndex
	{
		/// <summary>
		/// List of the filename extensions we categorize as photos.
		/// </summary>
		public static readonly string[] PhotoExtensions = { ".jpg", ".jpeg", ".png", ".gif" };

		/// <summary>
		/// List of the filename extensions we categorize as documents.
		/// </summary>
		public static readonly string[] DocumentExtensions = { ".pdf", ".doc", ".docx", ".xls", ".xlsx", ".txt" };

		private string _type;
		private string _description;

		[JsonConstructor]
		protected Attachment(Filename filename, Hash hash, int length) : base(hash, length)
		{
			Filename = filename;
		}

		public Attachment(Filename filename, [NotNull] byte[] data) : base(data)
		{
			Filename = filename;
		}

		[JsonIgnore]
		public int Index
		{
			get
			{
				var attachments = Parent as DocumentChildCollection<CrashReport, Attachment>;
				return attachments?.IndexOf(this) ?? -1;
			}
		}

		public override string AttachmentId => $"{Root?.Id:N}/attachments/{Hash}/{Filename}";

		public override Filename Filename { get; }

		[MaxLength(50)]
		public override string Type
		{
			get { return _type; }
			set
			{
				if (value == _type) return;
				_type = value;
				NotifyOfPropertyChange();
			}
		}

		public override string Description
		{
			get { return _description; }
			set
			{
				if (value == _description) return;
				_description = value;
				NotifyOfPropertyChange();
			}
		}

		public bool Equals(Attachment other)
		{
			if (ReferenceEquals(null, other)) return false;
			if (ReferenceEquals(this, other)) return true;
			return Hash.Equals(other.Hash) && Filename.Equals(other.Filename) && Length == other.Length;
		}

		public override bool Equals(object obj)
		{
			if (ReferenceEquals(null, obj)) return false;
			if (ReferenceEquals(this, obj)) return true;
			if (obj.GetType() != GetType()) return false;
			return Equals((Attachment)obj);
		}

		// ReSharper disable NonReadonlyFieldInGetHashCode
		public override int GetHashCode()
		{
			unchecked
			{
				var hashCode = Hash.GetHashCode();
				hashCode = (hashCode * 397) ^ Length;
				hashCode = (hashCode * 397) ^ Filename.GetHashCode();
				return hashCode;
			}
		}
		// ReSharper restore NonReadonlyFieldInGetHashCode

		public static bool operator ==(Attachment left, Attachment right)
		{
			return Equals(left, right);
		}

		public static bool operator !=(Attachment left, Attachment right)
		{
			return !Equals(left, right);
		}

		void INeedsCollectionIndex.OnCollectionIndexChanged()
		{
			NotifyOfPropertyChange(nameof(Index));
		}
	}
}