﻿using System.Text.RegularExpressions;
using ECrash.Core.Models;
using ECrash.Core.Models.Document;
using ECrash.Core.Models.ReferenceData;
using ECrashLA.Core.Models.CrashReports.ReferenceData;

namespace ECrashLA.Core.Models.CrashReports.BusinessRules;

public class DetermineFmcsaReportability : CrashReportBusinessRuleBase
{
	public DetermineFmcsaReportability(IReferenceDataCache referenceDataCache) : base(referenceDataCache, nameof(CrashReport.CrashSeverity))
	{
	}

	protected override void OnInitializeCore(CrashReport document, ReferenceDataFilter referenceData)
	{
		base.OnInitializeCore(document, referenceData);
		ApplyRule(document, referenceData, null);
	}

	protected override void OnNodeAttached(CrashReport document, IDocumentNode<CrashReport> node, ReferenceDataFilter referenceData)
	{
		base.OnNodeAttached(document, node, referenceData);
		if (node is Vehicle or Person)
		{
			ApplyRule(document, referenceData, null);
		}
	}

	protected override bool ShouldApplyRule(DocumentPropertyChangedEventArgs change)
	{
		return base.ShouldApplyRule(change) 
		       || (change.Source is Vehicle && change.IsForAny(nameof(Vehicle.TowStatus), 
		                                                       nameof(Vehicle.BodyType),
		                                                       nameof(Vehicle.HazardousMaterialsPlacardStatus),
		                                                       nameof(Vehicle.Vin),
		                                                       nameof(Vehicle.VehicleConfiguration),
		                                                       nameof(Vehicle.FmcsaReportability))) 
		       || (change.Source is Person && change.IsForAny(nameof(Person.InjuryStatus), 
		                                                      nameof(Person.MedicalTransportationType)));
	}

	protected override void ApplyRule(CrashReport document, ReferenceDataFilter referenceData, DocumentPropertyChangedEventArgs change)
	{
		ComputeVehicleFmcsaReportability(document);
		ComputeCrashFmcsaReportability(document);
	}

	private void ComputeVehicleFmcsaReportability(CrashReport document)
	{
		var referenceData = ReferenceDataCache.GetFiltered(document.ReferenceDataEpoch);
		foreach (var vehicle in document.Vehicles)
		{
			ComputeVehicleFmcsaReportability(vehicle, referenceData);
		}
	}

	private void ComputeVehicleFmcsaReportability(Vehicle vehicle, ReferenceDataFilter referenceData)
	{
		var reportability = VehicleFmcsaReportability.NotReportable;

		// Check to see if the vehicle is reportable due to its body type
		if (vehicle.BodyType?.IsLargeVehicle == true)
		{
			reportability |= VehicleFmcsaReportability.ReportableDueToBodyType;
		}

		// Check to see if the vehicle is reportable because it carried hazardous materials
		if (vehicle.HazardousMaterialsPlacardStatus?.IsYes == true)
		{
			reportability |= VehicleFmcsaReportability.ReportableDueToHazardousMaterials;
		}

		// Check to see if the vehicle's VIN matches patterns for known reportable vehicles
		if (!string.IsNullOrWhiteSpace(vehicle.Vin))
		{
			var rules = referenceData.Get<VehicleFmcsaReportabilityRule>();
			foreach (var rule in rules)
			{
				if (string.IsNullOrWhiteSpace(rule.VinRegex)) continue;
				if (Regex.IsMatch(vehicle.Vin, rule.VinRegex))
				{
					reportability |= rule.Reportability;
				}
			}
		}

		// Check to see if the vehicle's configuration is conditionally reportable
		if (vehicle.VehicleConfiguration?.IsYes == true)
		{
			reportability |= VehicleFmcsaReportability.ReportableDueToVehicleConfiguration;
		}

		vehicle.FmcsaReportability = reportability;
	}

	private static void ComputeCrashFmcsaReportability(CrashReport document)
	{
		var reportability = CrashFmcsaReportability.NotReportable;

		if (document.CrashSeverity?.IsFatal == true)
		{
			// Flag as reportable if the overall severity indicates a fatality
			reportability |= CrashFmcsaReportability.ReportableDueToFatality;
		}

		foreach (var vehicle in document.Vehicles)
		{
			if (vehicle.TowStatus?.IsDueToDisablingDamage == true)
			{
				// Flag as reportable if any vehicle was towed due to disabling damage
				reportability |= CrashFmcsaReportability.ReportableDueToTowaway;
			}
			if ((vehicle.FmcsaReportability & VehicleFmcsaReportability.ReportableMask) != 0)
			{
				// Flag if any vehicle is reportable
				reportability |= CrashFmcsaReportability.HasReportableVehicles;
			}
			if ((vehicle.FmcsaReportability & VehicleFmcsaReportability.ConditiallyReportableMask) != 0)
			{
				// Flag if any vehicle is conditionally reportable
				reportability |= CrashFmcsaReportability.HasConditionalReportableVehicles;
			}
		}

		foreach (var person in document.EnumeratePersons())
		{
			if (person.InjuryStatus?.IsFatal == true)
			{
				// Flag as reportable if any person was killed
				reportability |= CrashFmcsaReportability.ReportableDueToFatality;
			}
			else if (person.InjuryStatus?.IsInjury == true && person.MedicalTransportationType?.IsYes == true)
			{
				// Flag as reportable if any person was transported from the scene for medical attention
				reportability |= CrashFmcsaReportability.ReportableDueToInjury;
			}
		}

		document.FmcsaReportability = reportability;
	}
}