﻿using System.Linq;
using ECrash.Core.Models;
using ECrash.Core.Models.Document;
using ECrash.Core.Models.ReferenceData;
using ECrashLA.Core.Models.CrashReports.ReferenceData;

namespace ECrashLA.Core.Models.CrashReports.BusinessRules
{
	public class SetAlcoholInvolvement : CrashReportBusinessRuleBase
	{
		public const decimal BacThreshold = 0.02M;

		public SetAlcoholInvolvement(IReferenceDataCache referenceDataCache) : base(referenceDataCache)
		{
		}

		protected override void OnInitializeCore(CrashReport document, ReferenceDataFilter referenceData)
		{
			base.OnInitializeCore(document, referenceData);
			using (document.SuspendNotifications())
			{
				ApplyRule(document, referenceData, null);
			}
		}

		protected override bool ShouldApplyRule(DocumentPropertyChangedEventArgs change)
		{
			return change.Source is ConditionInfo && change.IsForAny(nameof(ConditionInfo.AlcoholUseSuspicion), nameof(ConditionInfo.AlcoholTestResult), nameof(ConditionInfo.BloodAlcoholContent));
		}

		protected override void OnNodeAttached(CrashReport document, IDocumentNode<CrashReport> node, ReferenceDataFilter referenceData)
		{
			base.OnNodeAttached(document, node, referenceData);
			if (node is Vehicle || node is Person || node is ConditionInfo)
			{
				ApplyRule(document, referenceData, null);
			}
		}

		protected override void ApplyRule(CrashReport document, ReferenceDataFilter referenceData, DocumentPropertyChangedEventArgs change)
		{
			bool isUnknown = false;
			bool isAlcoholInvolved = false;

			// See eCrashMS3-446 for details on this algorithm
			foreach (var conditionInfo in document.EnumeratePersons().Where(x => x is IHasConditionInfo).Select(x => ((IHasConditionInfo)x).ConditionInfo))
			{
				if ((conditionInfo.AlcoholTestResult != null && conditionInfo.AlcoholTestResult.IsNo && !conditionInfo.AlcoholTestResult.IsPending) || conditionInfo.BloodAlcoholContent < BacThreshold)
				{
					// Person tested negative for alcohol, so skip all other testing for this person
					continue;
				}
				if (conditionInfo.AlcoholUseSuspicion?.IsYes == true)
				{
					// Person suspected of alcohol use, so AlcoholInvovlement = true
					isAlcoholInvolved = true;
					break;
				}
				if (conditionInfo.AlcoholTestResult == null)
				{
					// Test result not entered, so AlcoholInvovlement = unknown unless true for another person
					isUnknown = true;
					continue;
				}
				if (conditionInfo.AlcoholTestResult.IsUnknown || conditionInfo.AlcoholTestResult.IsPending)
				{
					// Test result unknown or pending, so AlcoholInvovlement = unknown unless true for another person
					isUnknown = true;
					continue;
				}
				if (conditionInfo.AlcoholTestResult.IsYes != true) continue;
				if (conditionInfo.AlcoholTestResult.IsBloodAlcoholContentRequired)
				{
					if (conditionInfo.BloodAlcoholContent == null)
					{
						// BAC not entered, so AlcoholInvovlement = unknown unless true for another person
						isUnknown = true;
						continue;
					}
					// BAC over threshold (we checked for under threshold earlier), so AlcoholInvovlement = true
					isAlcoholInvolved = true;
					break;
				}
				// Positive results with no actual BAC value, so AlcoholInvovlement = true
				isAlcoholInvolved = true;
				break;
			}

			document.AlcoholInvolvement = isAlcoholInvolved 
				? referenceData.GetYes<AlcoholInvolvement>() 
				: isUnknown 
					? referenceData.GetUnknown<AlcoholInvolvement>() 
					: referenceData.GetNo<AlcoholInvolvement>();
		}
	}
}