﻿using System.Linq;
using ECrash.Core.Models;
using ECrash.Core.Models.Document;
using ECrash.Core.Models.ReferenceData;
using ECrashLA.Core.Models.CrashReports.ReferenceData;

namespace ECrashLA.Core.Models.CrashReports.BusinessRules
{
	public class SetDrugInvolvement : CrashReportBusinessRuleBase
	{
		public SetDrugInvolvement(IReferenceDataCache referenceDataCache) : base(referenceDataCache)
		{
		}

		protected override void OnInitializeCore(CrashReport document, ReferenceDataFilter referenceData)
		{
			base.OnInitializeCore(document, referenceData);
			using (document.SuspendNotifications())
			{
				ApplyRule(document, referenceData, null);
			}
		}

		protected override bool ShouldApplyRule(DocumentPropertyChangedEventArgs change)
		{
			return change.Source is ConditionInfo && change.IsFor(nameof(ConditionInfo.DrugUseSuspicion));
		}

		protected override void OnCollectionChangeCore(CrashReport document, ReferenceDataFilter referenceData, DocumentCollectionChangedEventArgs change)
		{
			base.OnCollectionChangeCore(document, referenceData, change);
			if (change.Source is DrugTestResultList)
			{
				ApplyRule(document, referenceData, null);
			}
		}

		protected override void OnNodeAttached(CrashReport document, IDocumentNode<CrashReport> node, ReferenceDataFilter referenceData)
		{
			base.OnNodeAttached(document, node, referenceData);
			if (node is Vehicle || node is Person || node is ConditionInfo)
			{
				ApplyRule(document, referenceData, null);
			}
		}

		// See eCrashMS3-446 for details on this algorithm
		protected override void ApplyRule(CrashReport document, ReferenceDataFilter referenceData, DocumentPropertyChangedEventArgs change)
		{
			bool isUnknown = false;
			bool drugsInvolved = false;

			foreach (var conditionInfo in document.EnumeratePersons().Where(x => x is IHasConditionInfo).Select(x => ((IHasConditionInfo)x).ConditionInfo))
			{
				// NOTE: we don't exclude people who tested negative in this algorithm because
				// they could be on a drug that wasn't tested for, according to Lt. Smith.
				if (conditionInfo.DrugUseSuspicion?.IsYes == true)
				{
					// Person suspected of drug use, so DrugInvolvement = true
					drugsInvolved = true;
					goto done;
				}
				// If no drug test results entered, DrugInvolvement = unknown unless true for another person
				isUnknown |= conditionInfo.DrugTestResults.Count == 0;
				foreach (var drugTestResult in conditionInfo.DrugTestResults)
				{
					if (drugTestResult.IsYes)
					{
						// Person tested positive for drug use, so DrugInvolvement = true
						drugsInvolved = true;
						goto done;
					}
					// Drug test result unknown for this person, so DrugInvolvement = unknown unless true for another person
					isUnknown |= drugTestResult.IsUnknown;
				}
			}

			done:
			document.DrugInvolvement = drugsInvolved
				? referenceData.GetYes<DrugInvolvement>() 
				: isUnknown 
					? referenceData.GetUnknown<DrugInvolvement>() 
					: referenceData.GetNo<DrugInvolvement>();
		}
	}
}