﻿using System;
using System.Buffers.Binary;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using ECrash.Core;
using ECrash.Core.Models;
using ECrash.Core.Models.Document;
using ECrash.Core.Models.JsonSchema;
using ECrash.Core.Models.ReferenceData;
using ECrashLA.Core.Models.CrashReports.ReferenceData;
using JetBrains.Annotations;
using Newtonsoft.Json;
using SimpleBase;
using ArgumentException = System.ArgumentException;

namespace ECrashLA.Core.Models.CrashReports
{
	public class CrashReport : DocumentRoot<CrashReport>, 
	                           IHasModifierSignature, 
	                           IHasStatus<CrashReportStatus>, 
	                           IHasRevisionNumber,
							   IHasUpdateTimestamp,
							   IHasKeywords
	{
		private CrashReportStatus _status;
		private DateTime _updateTimestamp;
		private Version _lastUpdatedWithVersion;
		private string _localCaseNumber;
		private string _localCaseReferenceNumber;
		private Guid _agencyId;
		private string _agencyName;
		private string _agencyCode;
		private string _agencyDivision;
		private string _investigatingOfficerRank;
		private string _investigatingOfficerFirstName;
		private string _investigatingOfficerMiddleName;
		private string _investigatingOfficerLastName;
		private string _investigatingOfficerNameSuffix;
		private string _investigatingOfficerId;
		private bool _photosTaken;
		private bool _videosTaken;
		private PropertyOwnership _propertyOwnership;
		private TrafficwayCharacteristics _trafficwayCharacteristics;
		private bool _secondaryCrash;
		private Date? _crashDate;
		private Time? _crashTime;
		private Date? _roadwayClearanceDate;
		private Time? _roadwayClearanceTime;
		private Date? _policeNotificationDate;
		private Time? _policeNotificationTime;
		private Date? _policeArrivalDate;
		private Time? _policeArrivalTime;
		private Date? _onSceneInvestigationCompletionDate;
		private Time? _onSceneInvestigationCompletionTime;
		private Parish _parish;
		private City _city;
		private decimal? _latitude;
		private decimal? _longitude;
		private RoadClassification _roadClassification;
		private RoadSubtype _roadSubtype;
		private LocationSource _locationSource;
		private string _troopRegion;
		private string _internalRouteId;
		private decimal? _milepoint;

		private Highway _highway;

		private string _roadNumber;
		private string _roadName;
		private string _roadType;

		private bool _isDistanceFromIntersectionNotApplicable;
		private decimal? _distanceFromIntersection;
		private DistanceUnit? _distanceFromIntersectionUnit;
		private DirectionFromIntersection _directionFromIntersection;

		private bool _isIntersection;

		private string _intersectingRoadName;
		private string _intersectingRoadType;

		private TrafficFlowDirection _trafficFlowDirection;
		private ContributingFactor _primaryContributingFactor;
		private ContributingFactor _secondaryContributingFactor;
		private VehicleEvent _firstHarmfulEvent;
		private string _firstHarmfulEventDescription;
		private VehicleEventLocation _firstHarmfulEventLocation;
		private CrashManner _crashManner;
		private string _crashMannerDescription;
		private string _weatherConditionsDescription;
		private LightCondition _lightCondition;
		private string _lightConditionDescription;
		private RoadwaySurfaceCondition _roadwaySurfaceCondition;
		private string _roadwaySurfaceConditionDescription;
		private string _roadwayEnvironmentalContributingCircumstancesDescription;
		private JunctionLocation _junctionLocation;
		private string _junctionLocationDescription;
		private IntersectionApproach _intersectionApproach;
		private IntersectionGeometry _intersectionGeometry;
		private IntersectionTrafficControl _intersectionTrafficControl;
		private SchoolBusRelation _schoolBusRelation;
		private WorkZoneRelation _workZoneRelation;
		private WorkZoneLocation _workZoneLocation;
		private WorkZoneType _workZoneType;
		private string _workZoneTypeDescription;
		private WorkZoneCircumstance _workZoneCircumstance;
		private string _workZoneCircumstanceDescription;
		private WorkerPresence _workerPresence;
		private LawEnforcementPresence _lawEnforcementPresence;
		private CrashSeverity _crashSeverity;
		private string _narrative;
		private string _officerNotes;
		private CrashFmcsaReportability _fmcsaReportability;

		// Minimum and maximum (exclusive) accepted values for latitude and longitude.
		public const decimal LatitudeMax = 33.019457m;
		public const decimal LatitudeMin = 28.928609m;
		public const decimal LongitudeMax = -88.817017m;
		public const decimal LongitudeMin = -94.043147m;

		[JsonConstructor]
		protected CrashReport(Guid id, Guid creatorUserId, [NotNull] string creatorUsername, DateTime creationTimestamp) : base(id)
		{
			if (creatorUsername == null) throw new ArgumentException("creatorUsername must not be null", nameof(creatorUsername));
			if (creatorUsername.Length > 100) throw new ArgumentException("creatorUsername must be 100 characters or less", nameof(creatorUsername));
			CreatorUserId = creatorUserId;
			CreatorUsername = creatorUsername;
			CreationTimestamp = creationTimestamp;

			// Compute human-readable ID
			ReadableId = ComputeHumanReadableId(id, creationTimestamp);

			WeatherConditions = new WeatherConditionList().InitParent(this);
			RoadwayEnvironmentalContributingCircumstances = new RoadwayEnvironmentalContributingCircumstanceList().InitParent(this);
			Vehicles = new VehicleList().InitParent(this);
			NonMotorists = new NonMotoristList().InitParent(this);
			Trains = new TrainList().InitParent(this);
			DamagedNonVehicularProperty = new NonVehicularPropertyList().InitParent(this);
			Witnesses = new WitnessList().InitParent(this);
			Attachments = new AttachmentList().InitParent(this);
			Diagrams = new DiagramList().InitParent(this);
		}

		public CrashReport(
			IUserIdentity creator,
			IList<IBusinessRule<CrashReport>> businessRules) : this(Guid.NewGuid(), creator.UserId, creator.Username, DateTime.Now)
		{
			_status = new CrashReportStatus(Id);
			UpdateTimestamp = CreationTimestamp;
			LastUpdatedWithVersion = Assembly.GetExecutingAssembly().GetName().Version;
			AttachBusinessRules(businessRules);
		}

		public override IEnumerable<IDocumentNode<CrashReport>> EnumerateChildNodes()
		{
			yield return WeatherConditions;
			yield return RoadwayEnvironmentalContributingCircumstances;
			yield return Vehicles;
			yield return NonMotorists;
			yield return Trains;
			yield return DamagedNonVehicularProperty;
			yield return Witnesses;
			yield return Attachments;
			yield return Diagrams;
		}

		[JsonIgnore]
		public CrashReportStatus Status
		{
			get => _status;
			set
			{
				_status = value;
				NotifyOfNonDocumentPropertyChange();
				NotifyOfNonDocumentPropertyChange(nameof(StateCaseNumber));
				NotifyOfNonDocumentPropertyChange(nameof(LatestRevision));
				if (_status != null)
				{
					// If the report has been updated since the last upload, update
					// the CrashReportStatus's copy of identifying fields
					if (UpdateTimestamp > _status.UploadTimestamp || _status.UploadTimestamp == null)
					{
						_status.LocalCaseNumber = LocalCaseNumber;
						_status.LocalCaseReferenceNumber = LocalCaseReferenceNumber;
						_status.Date = CrashDate;
						_status.Time = CrashTime;
					}
				}
			}
		}

		public Guid CreatorUserId { get; }
		[MaxLength(100)] public string CreatorUsername { get; }
		public DateTime CreationTimestamp { get; }

		public string ReadableId { get; }

		public DateTime UpdateTimestamp
		{
			get => _updateTimestamp;
			set => Set(ref _updateTimestamp, value);
		}

		public Version LastUpdatedWithVersion
		{
			get => _lastUpdatedWithVersion;
			set
			{
				Set(ref _lastUpdatedWithVersion, value);
				ReferenceDataEpoch = new ReferenceDataEpoch(ReferenceDataEffectiveTimestamp, _lastUpdatedWithVersion);
			}
		}

		[JsonIgnore]
		public override int LatestRevision
		{
			get => _status?.LatestRevision ?? 0;
			set
			{
				if (_status != null) _status.LatestRevision = value;
				NotifyOfNonDocumentPropertyChange();
			}
		}

		[MaxLength(50)]
		public string LocalCaseNumber
		{
			get => _localCaseNumber;
			set
			{
				Set(ref _localCaseNumber, value);
				if (Status != null) Status.LocalCaseNumber = _localCaseNumber;
			}
		}

		[MaxLength(50)]
		public string LocalCaseReferenceNumber
		{
			get => _localCaseReferenceNumber;
			set
			{
				Set(ref _localCaseReferenceNumber, value);
				if (Status != null) Status.LocalCaseReferenceNumber = _localCaseReferenceNumber;
			}
		}

		[MaxLength(50)]
		public string StateCaseNumber
		{
			get => _status?.StateCaseNumber;
			set
			{
				if (Status != null) Status.StateCaseNumber = value;
				NotifyOfNonDocumentPropertyChange();
			}
		}

		public Guid AgencyId
		{
			get => _agencyId;
			set => Set(ref _agencyId, value);
		}

		[MaxLength(200)]
		public string AgencyName
		{
			get => _agencyName;
			set => Set(ref _agencyName, value);
		}

		[MaxLength(100)]
		public string AgencyCode
		{
			get => _agencyCode;
			set => Set(ref _agencyCode, value);
		}

		[MaxLength(25)]
		public string AgencyDivision
		{
			get => _agencyDivision;
			set => Set(ref _agencyDivision, value);
		}

		[MaxLength(25)]
		public string InvestigatingOfficerRank
		{
			get => _investigatingOfficerRank;
			set => Set(ref _investigatingOfficerRank, value);
		}

		[MaxLength(50)]
		public string InvestigatingOfficerFirstName
		{
			get => _investigatingOfficerFirstName;
			set => Set(ref _investigatingOfficerFirstName, value);
		}

		[MaxLength(50)]
		public string InvestigatingOfficerMiddleName
		{
			get => _investigatingOfficerMiddleName;
			set => Set(ref _investigatingOfficerMiddleName, value);
		}

		[MaxLength(50)]
		public string InvestigatingOfficerLastName
		{
			get => _investigatingOfficerLastName;
			set => Set(ref _investigatingOfficerLastName, value);
		}

		[MaxLength(10)]
		public string InvestigatingOfficerNameSuffix
		{
			get => _investigatingOfficerNameSuffix;
			set => Set(ref _investigatingOfficerNameSuffix, value);
		}

		[MaxLength(12)]
		public string InvestigatingOfficerId
		{
			get => _investigatingOfficerId;
			set => Set(ref _investigatingOfficerId, value);
		}

		public bool PhotosTaken
		{
			get => _photosTaken;
			set => Set(ref _photosTaken, value);
		}

		public bool VideosTaken
		{
			get => _videosTaken;
			set => Set(ref _videosTaken, value);
		}

		public PropertyOwnership PropertyOwnership
		{
			get => _propertyOwnership;
			set => Set(ref _propertyOwnership, value);
		}

		public TrafficwayCharacteristics TrafficwayCharacteristics
		{
			get => _trafficwayCharacteristics;
			set => Set(ref _trafficwayCharacteristics, value);
		}

		/// <summary>
		/// Indicates whether the crash is a secondary crash, as defined by MMUCC field C2.3.
		/// This can include a vehicle traffic crash within a traffic incident scene or within a traffic queue resulting from a prior traffic incident.
		/// </summary>
		public bool SecondaryCrash
		{
			get => _secondaryCrash;
			set => Set(ref _secondaryCrash, value);
		}

		public Date? CrashDate
		{
			get => _crashDate;
			set
			{
				SetWithDependencies(ref _crashDate,
				                    value,
				                    nameof(CrashDate),
				                    nameof(CrashDateTime),
				                    nameof(ReferenceDataEffectiveTimestamp),
				                    nameof(TimeSpanBetweenCrashAndRoadwayClearance),
				                    nameof(TimeSpanBetweenCrashAndPoliceNotification)
				);
				if (Status != null) Status.Date = _crashDate;
				ReferenceDataEpoch = new ReferenceDataEpoch(ReferenceDataEffectiveTimestamp, _lastUpdatedWithVersion);
			}
		}

		public Time? CrashTime
		{
			get => _crashTime;
			set
			{
				SetWithDependencies(ref _crashTime,
				                    value,
				                    nameof(CrashTime),
				                    nameof(CrashDateTime),
				                    nameof(ReferenceDataEffectiveTimestamp),
				                    nameof(TimeSpanBetweenCrashAndRoadwayClearance),
				                    nameof(TimeSpanBetweenCrashAndPoliceNotification)
				);
				if (Status != null) Status.Time = _crashTime;
				ReferenceDataEpoch = new ReferenceDataEpoch(ReferenceDataEffectiveTimestamp, _lastUpdatedWithVersion);
			}
		}

		public DateTime? CrashDateTime => _crashDate.Combine(_crashTime);

		public DateTime ReferenceDataEffectiveTimestamp
		{
			get
			{
				if (_crashDate == null) return CreationTimestamp;
				if (CrashDateTime == null) return (DateTime)_crashDate.Value;
				return CrashDateTime.Value;
			}
		}

		public Date? RoadwayClearanceDate
		{
			get => _roadwayClearanceDate;
			set => SetWithDependencies(ref _roadwayClearanceDate, 
			                           value, 
			                           nameof(RoadwayClearanceDate), 
			                           nameof(RoadwayClearanceDateTime),
			                           nameof(TimeSpanBetweenCrashAndRoadwayClearance));
		}

		public Time? RoadwayClearanceTime
		{
			get => _roadwayClearanceTime;
			set => SetWithDependencies(ref _roadwayClearanceTime, 
			                           value, 
			                           nameof(RoadwayClearanceTime), 
			                           nameof(RoadwayClearanceDateTime),
			                           nameof(TimeSpanBetweenCrashAndRoadwayClearance));
		}

		public DateTime? RoadwayClearanceDateTime => _roadwayClearanceDate.Combine(_roadwayClearanceTime);

		[JsonIgnore]
		public TimeSpan? TimeSpanBetweenCrashAndRoadwayClearance
		{
			get
			{
				if (CrashDate == null || CrashTime == null || RoadwayClearanceDate == null || RoadwayClearanceTime == null) return null;
				return RoadwayClearanceDateTime - CrashDateTime;
			}
		}

		public Date? PoliceNotificationDate
		{
			get => _policeNotificationDate;
			set => SetWithDependencies(ref _policeNotificationDate,
			                           value,
			                           nameof(PoliceNotificationDate),
			                           nameof(PoliceNotificationDateTime),
			                           nameof(TimeSpanBetweenCrashAndPoliceNotification), 
			                           nameof(TimeSpanBetweenPoliceNotificationAndPoliceArrival));
		}

		public Time? PoliceNotificationTime
		{
			get => _policeNotificationTime;
			set => SetWithDependencies(ref _policeNotificationTime,
			                           value,
			                           nameof(PoliceNotificationTime),
			                           nameof(PoliceNotificationDateTime),
			                           nameof(TimeSpanBetweenCrashAndPoliceNotification),
			                           nameof(TimeSpanBetweenPoliceNotificationAndPoliceArrival));
		}

		public DateTime? PoliceNotificationDateTime => _policeNotificationDate.Combine(_policeNotificationTime);

		[JsonIgnore]
		public TimeSpan? TimeSpanBetweenCrashAndPoliceNotification
		{
			get
			{
				if (CrashDate == null || CrashTime == null || PoliceNotificationDate == null || PoliceNotificationTime == null) return null;
				return PoliceNotificationDateTime - CrashDateTime;
			}
		}

		public Date? PoliceArrivalDate
		{
			get => _policeArrivalDate;
			set => SetWithDependencies(ref _policeArrivalDate,
			           value,
			           nameof(PoliceArrivalDate),
			           nameof(PoliceArrivalDateTime),
			           nameof(TimeSpanBetweenPoliceNotificationAndPoliceArrival),
			           nameof(TimeSpanBetweenPoliceArrivalAndOnSceneInvestigationCompletion)
			);
		}

		public Time? PoliceArrivalTime
		{
			get => _policeArrivalTime;
			set => SetWithDependencies(ref _policeArrivalTime,
			           value,
			           nameof(PoliceArrivalTime),
			           nameof(PoliceArrivalDateTime),
			           nameof(TimeSpanBetweenPoliceNotificationAndPoliceArrival),
			           nameof(TimeSpanBetweenPoliceArrivalAndOnSceneInvestigationCompletion)
			);
		}

		public DateTime? PoliceArrivalDateTime => _policeArrivalDate.Combine(_policeArrivalTime);

		[JsonIgnore]
		public TimeSpan? TimeSpanBetweenPoliceNotificationAndPoliceArrival
		{
			get
			{
				if (PoliceNotificationDate == null || PoliceNotificationTime == null || PoliceArrivalDate == null || PoliceArrivalTime == null) return null;
				return PoliceArrivalDateTime - PoliceNotificationDateTime;
			}
		}

		public Date? OnSceneInvestigationCompletionDate
		{
			get => _onSceneInvestigationCompletionDate;
			set => SetWithDependencies(ref _onSceneInvestigationCompletionDate,
			           value,
			           nameof(OnSceneInvestigationCompletionDate),
			           nameof(OnSceneInvestigationCompletionDateTime),
			           nameof(TimeSpanBetweenPoliceArrivalAndOnSceneInvestigationCompletion)
			);
		}

		public Time? OnSceneInvestigationCompletionTime
		{
			get => _onSceneInvestigationCompletionTime;
			set => SetWithDependencies(ref _onSceneInvestigationCompletionTime,
			           value,
			           nameof(OnSceneInvestigationCompletionTime),
			           nameof(OnSceneInvestigationCompletionDateTime),
			           nameof(TimeSpanBetweenPoliceArrivalAndOnSceneInvestigationCompletion)
			);
		}

		public DateTime? OnSceneInvestigationCompletionDateTime => _onSceneInvestigationCompletionDate.Combine(_onSceneInvestigationCompletionTime);

		[JsonIgnore]
		public TimeSpan? TimeSpanBetweenPoliceArrivalAndOnSceneInvestigationCompletion
		{
			get
			{
				if (PoliceArrivalDate == null || PoliceArrivalTime == null || OnSceneInvestigationCompletionDate == null || OnSceneInvestigationCompletionTime == null) return null;
				return OnSceneInvestigationCompletionDateTime - PoliceArrivalDateTime;
			}
		}

		public Parish Parish
		{
			get => _parish;
			set => Set(ref _parish, value);
		}

		public City City
		{
			get => _city;
			set => Set(ref _city, value);
		}

		/// <summary>
		/// The crash location's latitude measured in degrees, as used in MMUCC field C6.
		/// </summary>
		public decimal? Latitude
		{
			get => _latitude;
			set => Set(ref _latitude, value);
		}

		/// <summary>
		/// The crash location's longitude measured in degrees, as used in MMUCC field C6.
		/// </summary>
		public decimal? Longitude
		{
			get => _longitude;
			set => Set(ref _longitude, value);
		}

		public RoadClassification RoadClassification
		{
			get => _roadClassification;
			set => Set(ref _roadClassification, value);
		}

		public RoadSubtype RoadSubtype
		{
			get => _roadSubtype;
			set => Set(ref _roadSubtype, value);
		}

		public LocationSource LocationSource
		{
			get => _locationSource;
			set => Set(ref _locationSource, value);
		}

		[MaxLength(2)]
		public string TroopRegion
		{
			get => _troopRegion;
			set => Set(ref _troopRegion, value);
		}

		[MaxLength(38)]
		public string InternalRouteId
		{
			get => _internalRouteId;
			set => Set(ref _internalRouteId, value);
		}

		public decimal? Milepoint
		{
			get => _milepoint;
			set => Set(ref _milepoint, value);
		}

		public Highway Highway
		{
			get => _highway;
			set => Set(ref _highway, value);
		}

		[MaxLength(10)]
		public string RoadNumber
		{
			get => _roadNumber;
			set => SetWithDependencies(ref _roadNumber, value, nameof(RoadNumber), nameof(FullRoadName));
		}

		[MaxLength(50)]
		public string RoadName
		{
			get => _roadName;
			set => SetWithDependencies(ref _roadName, value, nameof(RoadName), nameof(FullRoadName));
		}

		[MaxLength(15)]
		public string RoadType
		{
			get => _roadType;
			set => SetWithDependencies(ref _roadType, value, nameof(RoadType), nameof(FullRoadName));
		}

		[JsonIgnore]
		public string FullRoadName => " ".ImprovedJoin(RoadNumber, RoadName, RoadType);

		public bool IsDistanceFromIntersectionNotApplicable
		{
			get => _isDistanceFromIntersectionNotApplicable;
			set => Set(ref _isDistanceFromIntersectionNotApplicable, value);
		}

		public decimal? DistanceFromIntersection
		{
			get => _distanceFromIntersection;
			set => Set(ref _distanceFromIntersection, value);
		}

		public DistanceUnit? DistanceFromIntersectionUnit
		{
			get => _distanceFromIntersectionUnit;
			set => Set(ref _distanceFromIntersectionUnit, value);
		}

		public DirectionFromIntersection DirectionFromIntersection
		{
			get => _directionFromIntersection;
			set => Set(ref _directionFromIntersection, value);
		}

		public TrafficFlowDirection TrafficFlowDirection
		{
			get => _trafficFlowDirection;
			set => Set(ref _trafficFlowDirection, value);
		}

		public bool IsIntersection
		{
			get => _isIntersection;
			set => Set(ref _isIntersection, value);
		}

		[MaxLength(50)]
		public string IntersectingRoadName
		{
			get => _intersectingRoadName;
			set => SetWithDependencies(ref _intersectingRoadName, value, nameof(IntersectingRoadName), nameof(FullIntersectingRoadName));
		}

		[MaxLength(15)]
		public string IntersectingRoadType
		{
			get => _intersectingRoadType;
			set => SetWithDependencies(ref _intersectingRoadType, value, nameof(IntersectingRoadType), nameof(FullIntersectingRoadName));
		}

		[JsonIgnore]
		public string FullIntersectingRoadName => " ".ImprovedJoin(IntersectingRoadName, IntersectingRoadType);

		public ContributingFactor PrimaryContributingFactor
		{
			get => _primaryContributingFactor;
			set => Set(ref _primaryContributingFactor, value);
		}

		public ContributingFactor SecondaryContributingFactor
		{
			get => _secondaryContributingFactor;
			set => Set(ref _secondaryContributingFactor, value);
		}

		public VehicleEvent FirstHarmfulEvent
		{
			get => _firstHarmfulEvent;
			set => Set(ref _firstHarmfulEvent, value);
		}

		[MaxLength(200)]
		public string FirstHarmfulEventDescription
		{
			get => _firstHarmfulEventDescription;
			set => Set(ref _firstHarmfulEventDescription, value);
		}

		public VehicleEventLocation FirstHarmfulEventLocation
		{
			get => _firstHarmfulEventLocation;
			set => Set(ref _firstHarmfulEventLocation, value);
		}

		public CrashManner CrashManner
		{
			get => _crashManner;
			set => Set(ref _crashManner, value);
		}

		[MaxLength(200)]
		public string CrashMannerDescription
		{
			get => _crashMannerDescription;
			set => Set(ref _crashMannerDescription, value);
		}

		[ListCannotContainNull]
		public WeatherConditionList WeatherConditions { get; }

		[MaxLength(200)]
		public string WeatherConditionsDescription
		{
			get => _weatherConditionsDescription;
			set => Set(ref _weatherConditionsDescription, value);
		}

		public LightCondition LightCondition
		{
			get => _lightCondition;
			set => Set(ref _lightCondition, value);
		}

		[MaxLength(200)]
		public string LightConditionDescription
		{
			get => _lightConditionDescription;
			set => Set(ref _lightConditionDescription, value);
		}

		public RoadwaySurfaceCondition RoadwaySurfaceCondition
		{
			get => _roadwaySurfaceCondition;
			set => Set(ref _roadwaySurfaceCondition, value);
		}

		[MaxLength(200)]
		public string RoadwaySurfaceConditionDescription
		{
			get => _roadwaySurfaceConditionDescription;
			set => Set(ref _roadwaySurfaceConditionDescription, value);
		}

		[ListCannotContainNull] 
		public RoadwayEnvironmentalContributingCircumstanceList RoadwayEnvironmentalContributingCircumstances { get; }

		[MaxLength(200)]
		public string RoadwayEnvironmentalContributingCircumstancesDescription
		{
			get => _roadwayEnvironmentalContributingCircumstancesDescription;
			set => Set(ref _roadwayEnvironmentalContributingCircumstancesDescription, value);
		}

		public JunctionLocation JunctionLocation
		{
			get => _junctionLocation;
			set => Set(ref _junctionLocation, value);
		}

		[MaxLength(200)]
		public string JunctionLocationDescription
		{
			get => _junctionLocationDescription;
			set => Set(ref _junctionLocationDescription, value);
		}

		public IntersectionApproach IntersectionApproach
		{
			get => _intersectionApproach;
			set => Set(ref _intersectionApproach, value);
		}

		public IntersectionGeometry IntersectionGeometry
		{
			get => _intersectionGeometry;
			set => Set(ref _intersectionGeometry, value);
		}

		public IntersectionTrafficControl IntersectionTrafficControl
		{
			get => _intersectionTrafficControl;
			set => Set(ref _intersectionTrafficControl, value);
		}

		public SchoolBusRelation SchoolBusRelation
		{
			get => _schoolBusRelation;
			set => Set(ref _schoolBusRelation, value);
		}

		public WorkZoneRelation WorkZoneRelation
		{
			get => _workZoneRelation;
			set => Set(ref _workZoneRelation, value);
		}

		public WorkZoneLocation WorkZoneLocation
		{
			get => _workZoneLocation;
			set => Set(ref _workZoneLocation, value);
		}

		public WorkZoneType WorkZoneType
		{
			get => _workZoneType;
			set => Set(ref _workZoneType, value);
		}

		[MaxLength(200)]
		public string WorkZoneTypeDescription
		{
			get => _workZoneTypeDescription;
			set => Set(ref _workZoneTypeDescription, value);
		}

		public WorkZoneCircumstance WorkZoneCircumstance
		{
			get => _workZoneCircumstance;
			set => Set(ref _workZoneCircumstance, value);
		}

		[MaxLength(200)]
		public string WorkZoneCircumstanceDescription
		{
			get => _workZoneCircumstanceDescription;
			set => Set(ref _workZoneCircumstanceDescription, value);
		}

		public WorkerPresence WorkerPresence
		{
			get => _workerPresence;
			set => Set(ref _workerPresence, value);
		}

		public LawEnforcementPresence LawEnforcementPresence
		{
			get => _lawEnforcementPresence;
			set => Set(ref _lawEnforcementPresence, value);
		}

		public CrashSeverity CrashSeverity
		{
			get => _crashSeverity;
			set => Set(ref _crashSeverity, value);
		}

		// These two are computed by business rules, so I don't bother with change notification on them
		public AlcoholInvolvement AlcoholInvolvement { get; set; }
		public DrugInvolvement DrugInvolvement { get; set; }

		[ListCannotContainNull]
		public VehicleList Vehicles { get; }
		[ListCannotContainNull]
		public NonMotoristList NonMotorists { get; }
		[ListCannotContainNull]
		public TrainList Trains { get; }
		[ListCannotContainNull]
		public NonVehicularPropertyList DamagedNonVehicularProperty { get; }
		[ListCannotContainNull]
		public WitnessList Witnesses { get; }
		[ListCannotContainNull]
		public AttachmentList Attachments { get; }
		[ListCannotContainNull]
		public DiagramList Diagrams { get; }

		public string Narrative
		{
			get => _narrative;
			set => Set(ref _narrative, value);
		}

		public string OfficerNotes
		{
			get => _officerNotes;
			set => Set(ref _officerNotes, value);
		}

		public CrashFmcsaReportability FmcsaReportability
		{
			get => _fmcsaReportability;
			set => Set(ref _fmcsaReportability, value);
		}

		[MaxLength(200)]
		[JsonSchemaNotNullable]
		public string ModifierSignature { get; set; }

		[JsonSchemaNotNullable]
		public DateTime? ModifierSignatureTimestamp { get; set; }

		public string GetKeywords()
		{
			var sb = new StringBuilder();

			sb.AppendLine(LocalCaseNumber);
			sb.AppendLine(LocalCaseReferenceNumber);
			sb.AppendLine(Status?.StateCaseNumber);
			sb.Append(RoadNumber).Append(' ').Append(RoadName).Append(' ').AppendLine(RoadType);
			sb.Append(IntersectingRoadName).Append(' ').AppendLine(IntersectingRoadType);

			foreach (var vehicle in Vehicles)
			{
				sb.AppendLine(vehicle.Vin);
				sb.AppendLine(vehicle.TagNumber);
				sb.AppendLine(vehicle.Make?.Name);
				sb.AppendLine(vehicle.Model?.Name);
				sb.AppendLine(vehicle.OwnerName);
				sb.AppendLine(vehicle.MotorCarrierName);

				foreach (var trailer in vehicle.Trailers)
				{
					sb.AppendLine(trailer.Vin);
					sb.AppendLine(trailer.TagNumber);
				}

				var driver = vehicle.Driver;
				if (driver != null)
				{
					sb.AppendLine(driver.FirstName);
					sb.AppendLine(driver.MiddleName);
					sb.AppendLine(driver.LastName);
					sb.AppendLine(driver.DateOfBirth?.ToIsoString());
					sb.AppendLine(driver.DriverLicenseNumber);

					foreach (var citation in driver.Citations)
					{
						sb.AppendLine(citation.TicketNumber);
					}
				}

				foreach (var passenger in vehicle.Passengers)
				{
					sb.AppendLine(passenger.FirstName);
					sb.AppendLine(passenger.MiddleName);
					sb.AppendLine(passenger.LastName);
					sb.AppendLine(passenger.DateOfBirth?.ToIsoString());
				}
			}

			foreach (var nonMotorist in NonMotorists)
			{
				sb.AppendLine(nonMotorist.FirstName);
				sb.AppendLine(nonMotorist.MiddleName);
				sb.AppendLine(nonMotorist.LastName);
				sb.AppendLine(nonMotorist.DateOfBirth?.ToIsoString());
			}

			foreach (var train in Trains)
			{
				sb.AppendLine(train.OperatorName);
				sb.AppendLine(train.DotCrossingNumber);

				foreach (var passenger in train.Passengers)
				{
					sb.AppendLine(passenger.FirstName);
					sb.AppendLine(passenger.MiddleName);
					sb.AppendLine(passenger.LastName);
					sb.AppendLine(passenger.DateOfBirth?.ToIsoString());
				}

				if (train.Conductor != null)
                {
					sb.AppendLine(train.Conductor.FirstName);
					sb.AppendLine(train.Conductor.MiddleName);
					sb.AppendLine(train.Conductor.LastName);
					sb.AppendLine(train.Conductor.DateOfBirth?.ToIsoString());
				}

				if (train.Engineer != null)
				{
					sb.AppendLine(train.Engineer.FirstName);
					sb.AppendLine(train.Engineer.MiddleName);
					sb.AppendLine(train.Engineer.LastName);
					sb.AppendLine(train.Engineer.DateOfBirth?.ToIsoString());
				}
			}

			foreach (var witness in Witnesses)
			{
				sb.AppendLine(witness.FirstName);
				sb.AppendLine(witness.MiddleName);
				sb.AppendLine(witness.LastName);
			}

			return sb.ToString();
		}

		public override IEnumerable<AttachmentBase<CrashReport>> EnumerateAttachments(Predicate<AttachmentBase<CrashReport>> includePredicate = null)
		{
			includePredicate ??= (_ => true);
			foreach (var attachment in Attachments)
			{
				if (includePredicate(attachment)) yield return attachment;
			}
			foreach (var diagram in Diagrams)
			{
				if (includePredicate(diagram)) yield return diagram;
			}
		}

		public IEnumerable<Person> EnumeratePersons()
		{
			foreach (var vehicle in Vehicles)
			{
				if (vehicle.Driver != null) yield return vehicle.Driver;
				foreach (var passenger in vehicle.Passengers)
				{
					yield return passenger;
				}
			}
			foreach (var nonMotorist in NonMotorists)
			{
				yield return nonMotorist;
			}
			foreach (var train in Trains)
			{
				if (train.Engineer != null)
				{
					yield return train.Engineer;
				}
				if (train.Conductor != null)
				{
					yield return train.Conductor;
				}
				foreach (var passenger in train.Passengers)
				{
					yield return passenger;
				}
			}
		}

		public IEnumerable<ICrashUnit> EnumerateUnits() => Vehicles.Cast<ICrashUnit>().Union(NonMotorists).Union(Trains);

		private static string ComputeHumanReadableId(Guid id, DateTime creationTimestamp)
		{
			return $"{creationTimestamp:yyMMdd}-{creationTimestamp:HHmmss}-{id.ToString("N")[..4].ToUpperInvariant()}";
		}
	}
}