﻿using System;
using System.Collections.Generic;
using ECrash.Core.Models;
using ECrash.Core.Models.Document;
using ECrash.Core.Models.Utilities;
using ECrashLA.Core.Models.CrashReports.ReferenceData;
using Newtonsoft.Json;

namespace ECrashLA.Core.Models.CrashReports
{
	public abstract class Person : DocumentChild<CrashReport>, INeedsCollectionIndex
	{
		private string _lastName;
		private string _firstName;
		private string _middleName;
		private string _nameSuffix;
		private bool _isNameUnknown;
		private bool _isAddressUnknown;
		private string _addressStreet;
		private string _addressCity;
		private State _addressState;
		private string _addressStateDescription;
		private string _addressPostalCode;
		private bool _isPhoneNumberUnknown;
		private string _phoneNumber;
		private Date? _dateOfBirth;
		private bool _isDateOfBirthUnknown;
		private int? _age;
		private bool _isAgeUnknown;
		private Sex _sex;
		private Race _race;
		private string _raceDescription;
		private Ethnicity _ethnicity;
		private IncidentResponder _incidentResponder;
		private string _incidentResponderDescription;
		private InjuryStatus _injuryStatus;
		private MedicalTransportationType _medicalTransportationType;
		private string _medicalTransportationTypeDescription;
		private EmergencyMedicalService _emsResponseAgency;
		private string _emsResponseAgencyDescription;
		private string _emsResponseRunNumber;
		private bool _isEmsResponseRunNumberUnknown;
		private MedicalFacility _medicalFacilityReceivingPatient;
		private string _medicalFacilityReceivingPatientDescription;
		private bool _isMedicalUniqueIdentifierUnknown;
		private bool _isMedicalUniqueIdentifierNotApplicable;
		private string _medicalUniqueIdentifier;

		/// <summary>
		/// Persons whose age is lower than this have the result values of their alcohol and drug tests hidden or redacted.
		/// </summary>
		public const int MinimumAgeForAlcoholAndDrugResults = 18;

		public override IEnumerable<IDocumentNode<CrashReport>> EnumerateChildNodes()
		{
			yield break;
		}

		[JsonIgnore]
		public abstract int Index { get; }

		[JsonIgnore]
		public int Number => Index + 1;

		[MaxLength(50)]
		public string LastName
		{
			get => _lastName;
			set => SetWithDependencies(ref _lastName, value, nameof(LastName), nameof(FullName));
		}

		[MaxLength(50)]
		public string FirstName
		{
			get => _firstName;
			set => SetWithDependencies(ref _firstName, value, nameof(FirstName), nameof(FullName));
		}

		[MaxLength(50)]
		public string MiddleName
		{
			get => _middleName;
			set => SetWithDependencies(ref _middleName, value, nameof(MiddleName), nameof(FullName));
		}

		[MaxLength(10)]
		public string NameSuffix
		{
			get => _nameSuffix;
			set => SetWithDependencies(ref _nameSuffix, value, nameof(NameSuffix), nameof(FullName));
		}

		[JsonIgnore]
		public string FullName => GetFullName(PersonNameOrder.FirstNameFirst);

		public bool IsNameUnknown
		{
			get => _isNameUnknown;
			set => Set(ref _isNameUnknown, value);
		}

		public bool IsAddressUnknown
		{
			get => _isAddressUnknown;
			set => Set(ref _isAddressUnknown, value);
		}
		
		[MaxLength(150)]
		public string AddressStreet
		{
			get => _addressStreet;
			set => Set(ref _addressStreet, value);
		}

		[MaxLength(50)]
		public string AddressCity
		{
			get => _addressCity;
			set => Set(ref _addressCity, value);
		}

		public State AddressState
		{
			get => _addressState;
			set => Set(ref _addressState, value);
		}

		[MaxLength(200)]
		public string AddressStateDescription
		{
			get => _addressStateDescription;
			set => Set(ref _addressStateDescription, value);
		}

		[MaxLength(20)]
		public string AddressPostalCode
		{
			get => _addressPostalCode;
			set => Set(ref _addressPostalCode, value);
		}

		public bool IsPhoneNumberUnknown
		{
			get => _isPhoneNumberUnknown;
			set => Set(ref _isPhoneNumberUnknown, value);
		}

		[MaxLength(20)]
		public string PhoneNumber
		{
			get => _phoneNumber;
			set => Set(ref _phoneNumber, value);
		}

		public Date? DateOfBirth
		{
			get => _dateOfBirth;
			set => Set(ref _dateOfBirth, value);
		}

		public bool IsDateOfBirthUnknown
		{
			get => _isDateOfBirthUnknown;
			set => Set(ref _isDateOfBirthUnknown, value);
		}

		public int? Age
		{
			get => _age;
			set => Set(ref _age, value);
		}

		public bool IsAgeUnknown
		{
			get => _isAgeUnknown;
			set => Set(ref _isAgeUnknown, value);
		}

		public Sex Sex
		{
			get => _sex;
			set => Set(ref _sex, value);
		}

		public Race Race
		{
			get => _race;
			set => Set(ref _race, value);
		}

		[MaxLength(200)]
		public string RaceDescription
		{
			get => _raceDescription;
			set => Set(ref _raceDescription, value);
		}

		public Ethnicity Ethnicity
		{
			get => _ethnicity;
			set => Set(ref _ethnicity, value);
		}

		public IncidentResponder IncidentResponder
		{
			get => _incidentResponder;
			set => Set(ref _incidentResponder, value);
		}

		[MaxLength(200)]
		public string IncidentResponderDescription
		{
			get => _incidentResponderDescription;
			set => Set(ref _incidentResponderDescription, value);
		}

		public InjuryStatus InjuryStatus
		{
			get => _injuryStatus;
			set => Set(ref _injuryStatus, value);
		}

		public MedicalTransportationType MedicalTransportationType
		{
			get => _medicalTransportationType;
			set => Set(ref _medicalTransportationType, value);
		}

		[MaxLength(200)]
		public string MedicalTransportationTypeDescription
		{
			get => _medicalTransportationTypeDescription;
			set => Set(ref _medicalTransportationTypeDescription, value);
		}

		public EmergencyMedicalService EmsResponseAgency
		{
			get => _emsResponseAgency;
			set => Set(ref _emsResponseAgency, value);
		}

		[MaxLength(200)]
		public string EmsResponseAgencyDescription
		{
			get => _emsResponseAgencyDescription;
			set => Set(ref _emsResponseAgencyDescription, value);
		}

		[MaxLength(20)]
		public string EmsResponseRunNumber
		{
			get => _emsResponseRunNumber;
			set => Set(ref _emsResponseRunNumber, value);
		}

		public bool IsEmsResponseRunNumberUnknown
		{
			get => _isEmsResponseRunNumberUnknown;
			set => Set(ref _isEmsResponseRunNumberUnknown, value);
		}

		public MedicalFacility MedicalFacilityReceivingPatient
		{
			get => _medicalFacilityReceivingPatient;
			set => Set(ref _medicalFacilityReceivingPatient, value);
		}

		[MaxLength(200)]
		public string MedicalFacilityReceivingPatientDescription
		{
			get => _medicalFacilityReceivingPatientDescription;
			set => Set(ref _medicalFacilityReceivingPatientDescription, value);
		}

		public bool IsMedicalUniqueIdentifierUnknown
		{
			get => _isMedicalUniqueIdentifierUnknown;
			set => Set(ref _isMedicalUniqueIdentifierUnknown, value);
		}

		public bool IsMedicalUniqueIdentifierNotApplicable
		{
			get => _isMedicalUniqueIdentifierNotApplicable;
			set => Set(ref _isMedicalUniqueIdentifierNotApplicable, value);
		}

		[MaxLength(50)]
		public string MedicalUniqueIdentifier
		{
			get => _medicalUniqueIdentifier;
			set => Set(ref _medicalUniqueIdentifier, value);
		}

		void INeedsCollectionIndex.OnCollectionIndexChanged()
		{
			NotifyOfPropertyChange(nameof(Index));
			NotifyOfPropertyChange(nameof(Number));
		}

		public string GetFullName(PersonNameOrder nameOrder)
		{
			return NameUtilities.GetFullName(nameOrder, _firstName, _middleName, _lastName, _nameSuffix);
		}

		public int? ComputeAgeAsOf(Date? asOfDate)
		{
			if (asOfDate == null || _dateOfBirth == null) return null;
			var asOfDateTime = (DateTime)asOfDate.Value;
			var dateOfBirth = (DateTime)_dateOfBirth.Value;
			// Algorithm for calculating age from http://stackoverflow.com/a/1404/156065
			int age = asOfDateTime.Year - dateOfBirth.Year;
			if (dateOfBirth > asOfDateTime.AddYears(-age)) age--;
			return age;
		}
	}
}