﻿using System;
using System.Collections.Generic;
using System.Linq;
using ECrash.Core;
using ECrash.Core.Models.ReferenceData;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace ECrashLA.Core.Models.CrashReports.ReferenceData
{
	public class VehicleModel : ReferenceDataBase
	{
		public const string MakeField = nameof(Make);
		public const string MakeCodeField = nameof(MakeCode);
		public const string FarsCodesField = nameof(FarsCodes);
		public const string NhtsaValueIdsField = nameof(NhtsaValueIds);
		public const string AbbreviationField = nameof(Abbreviation);

		public VehicleModel(int id, RowVersion version) : base(id, version)
		{
		}

		[JsonIgnore]
		public string Make
		{
			get => GetFieldValue<string>(MakeField);
			set => Fields[MakeField] = value;
		}

		[JsonIgnore]
		public string MakeCode
		{
			get => GetFieldValue<string>(MakeCodeField);
			set => Fields[MakeCodeField] = value;
		}

		[JsonIgnore]
		public IEnumerable<string> FarsCodes
		{
			get => GetFieldValues<string>(FarsCodesField);
			set => Fields[FarsCodesField] = JArray.FromObject(value ?? Array.Empty<string>());
		}

		public bool IsModelOf(VehicleMake make) => make?.Code == MakeCode;

		/// <summary>
		/// Value IDs used by the National Highway Traffic Safety Administration for this model.
		/// Used in VIN imports.
		/// </summary>
		[JsonIgnore]
		public IEnumerable<string> NhtsaValueIds
		{
			get => GetFieldValues<string>(NhtsaValueIdsField);
			set => Fields[NhtsaValueIdsField] = JArray.FromObject(value ?? Array.Empty<string>());
		}

		/// <summary>
		/// Abbreviations from NCIC that are matched to this VehicleModel.
		/// "Abbreviation" property name comes from the same property in VehicleMake.
		/// </summary>
		[JsonIgnore]
		public IEnumerable<string> Abbreviation
		{
			get => GetFieldValues<string>(AbbreviationField);
			set => Fields[AbbreviationField] = JArray.FromObject(value ?? Array.Empty<string>());
		}

		/// <summary>
		/// Returns true if <paramref name="model"/> matches this model's name or one of its abbreviations.
		/// </summary>
		/// <param name="model"></param>
		/// <returns></returns>
		public bool IsMatchByNameOrAbbreviation(string model)
		{
			if (string.IsNullOrWhiteSpace(model))
			{
				return false;
			}

			if (Name.Equals(model, StringComparison.InvariantCultureIgnoreCase))
			{
				return true;
			}

			if (Abbreviation?.Any(x => x.Replace('0', 'O').Equals(model.Replace('0', 'O'), StringComparison.InvariantCultureIgnoreCase)) == true)
			{
				return true;
			}

			return false;
		}
	}
}