﻿using System.Collections.Generic;
using ECrash.Core.Models.Document;
using ECrash.Core.Models.JsonSchema;
using ECrashLA.Core.Models.CrashReports.ReferenceData;
using Newtonsoft.Json;

namespace ECrashLA.Core.Models.CrashReports
{
	public class Train : DocumentChild<CrashReport>, INeedsCollectionIndex, ICrashUnit
	{
		private TrainType _trainType;

		private bool _isDotCrossingNumberNotApplicable;
		private bool _isDotCrossingNumberUnknown;
		private string _dotCrossingNumber;

		private TrainCrossingType _crossingType;
		private TrainCrossingSurfaceMaterial _crossingSurfaceMaterial;
		private string _crossingSurfaceMaterialDescription;
		private int? _setsOfTracks;
		private int? _trackSpeedLimit;
		private TrainMotion _motion;
		private int? _estimatedSpeedBeforeBraking;

		private bool _isTrainIdNumberUnknown;
		private bool _isTrainIdNumberNotApplicable;
		private string _trainIdNumber;

		private bool _isOperatorNameUnknown;
		private string _operatorName;
		private bool _isOperatorAddressUnknown;
		private string _operatorAddressStreet;
		private string _operatorAddressCity;
		private State _operatorAddressState;
		private string _operatorAddressStateDescription;
		private string _operatorAddressPostalCode;

		private bool _isTrackOwnerNameUnknown;
		private string _trackOwnerName;
		private bool _isTrackOwnerAddressUnknown;
		private string _trackOwnerAddressStreet;
		private string _trackOwnerAddressCity;
		private State _trackOwnerAddressState;
		private string _trackOwnerAddressStateDescription;
		private string _trackOwnerAddressPostalCode;

		private string _warningDevicesDescription;
		private string _advanceWarningDevicesDescription;
		private string _activeWarningDeviceStatusesDescription;

		private TrainCrossingVehicleInteraction _crossingVehicleInteraction;

		private bool _isMakeUnknown;
		private string _make;

		private bool _isTypeUnknown;
		private string _type;

		private bool _isLeadEngineNumberUnknown;
		private string _leadEngineNumber;

		private bool _isSerialNumberUnknown;
		private string _serialNumber;

		private bool _isNumberOfEnginesUnknown;
		private int? _numberOfEngines;

		private bool _isNumberOfCarsUnknown;
		private int? _numberOfCars;

		private bool _isDistanceTraveledAfterImpactNotApplicable;
		private decimal? _distanceTraveledAfterImpact;
		private DistanceUnit? _distanceTraveledAfterImpactUnit;

		private bool _isDataRecorderSpeedPending;
		private int? _dataRecorderSpeed;

		private TrainCollisionType _collisionType;

		private bool _isStruckCarPositionNotApplicable;
		private bool _isStruckCarPositionUnknown;
		private int? _struckCarPosition;

		private bool _isStruckCarTypeNotApplicable;
		private bool _isStruckCarTypeUnknown;
		private string _struckCarType;

		private bool _isStruckCarNumberNotApplicable;
		private bool _isStruckCarNumberUnknown;
		private string _struckCarNumber;

		private HazardousMaterialsPlacardStatus _hazardousMaterialsPlacardStatus;
		private bool _isHazardousMaterialsIdUnknown;
		private bool _isHazardousMaterialsIdNotApplicable;
		private string _hazardousMaterialsId;
		private HazardousMaterialClass _hazardousMaterialClass;
		private HazardousMaterialRelease _hazardousMaterialRelease;


		private Engineer _engineer;
		private Conductor _conductor;

		public Train()
		{
			WarningDevices = new TrainWarningDeviceList().InitParent(this);
			AdvanceWarningDevices = new TrainAdvanceWarningDeviceList().InitParent(this);
			ActiveWarningDeviceStatuses = new TrainActiveWarningDeviceStatusList().InitParent(this);
			EquipmentStatuses = new TrainEquipmentStatusList().InitParent(this);
			Engineer = new Engineer();
			Conductor = new Conductor();
			Passengers = new TrainPassengerList().InitParent(this);
		}

		public override IEnumerable<IDocumentNode<CrashReport>> EnumerateChildNodes()
		{
			yield return WarningDevices;
			yield return AdvanceWarningDevices;
			yield return ActiveWarningDeviceStatuses;
			yield return EquipmentStatuses;
			if (Engineer != null) yield return Engineer;
			if (Conductor != null) yield return Conductor;
			yield return Passengers;
		}

		/// <summary>
		/// Gets or sets the zero-based index of this vehicle.
		/// </summary>
		[JsonIgnore]
		public int Index => (Parent as TrainList)?.IndexOf(this) ?? -1;

		/// <summary>
		/// Gets the one-based number of this vehicle.
		/// </summary>
		[JsonIgnore]
		public int Number => Index + 1;

		public TrainType TrainType
		{
			get => _trainType;
			set => Set(ref _trainType, value);
		}

		public bool IsDotCrossingNumberNotApplicable
		{
			get => _isDotCrossingNumberNotApplicable;
			set => Set(ref _isDotCrossingNumberNotApplicable, value);
		}

		public bool IsDotCrossingNumberUnknown
		{
			get => _isDotCrossingNumberUnknown;
			set => Set(ref _isDotCrossingNumberUnknown, value);
		}

		[MaxLength(10)]
		public string DotCrossingNumber
		{
			get => _dotCrossingNumber;
			set => Set(ref _dotCrossingNumber, value);
		}

		public TrainCrossingType CrossingType
		{
			get => _crossingType;
			set => Set(ref _crossingType, value);
		}

		public TrainCrossingSurfaceMaterial CrossingSurfaceMaterial
		{
			get => _crossingSurfaceMaterial;
			set => Set(ref _crossingSurfaceMaterial, value);
		}

		[MaxLength(200)]
		public string CrossingSurfaceMaterialDescription
		{
			get => _crossingSurfaceMaterialDescription;
			set => Set(ref _crossingSurfaceMaterialDescription, value);
		}

		public int? SetsOfTracks
		{
			get => _setsOfTracks;
			set => Set(ref _setsOfTracks, value);
		}

		public int? TrackSpeedLimit
		{
			get => _trackSpeedLimit;
			set => Set(ref _trackSpeedLimit, value);
		}

		public TrainMotion Motion
		{
			get => _motion;
			set => Set(ref _motion, value);
		}

		public int? EstimatedSpeedBeforeBraking
		{
			get => _estimatedSpeedBeforeBraking;
			set => Set(ref _estimatedSpeedBeforeBraking, value);
		}

		public bool IsTrainIdNumberUnknown
		{
			get => _isTrainIdNumberUnknown;
			set => Set(ref _isTrainIdNumberUnknown, value);
		}

		public bool IsTrainIdNumberNotApplicable
		{
			get => _isTrainIdNumberNotApplicable;
			set => Set(ref _isTrainIdNumberNotApplicable, value);
		}

		[MaxLength(20)]
		public string TrainIdNumber
		{
			get => _trainIdNumber;
			set => Set(ref _trainIdNumber, value);
		}

		public bool IsOperatorNameUnknown
		{
			get => _isOperatorNameUnknown;
			set => Set(ref _isOperatorNameUnknown, value);
		}

		[MaxLength(50)]
		public string OperatorName
		{
			get => _operatorName;
			set => Set(ref _operatorName, value);
		}

		public bool IsOperatorAddressUnknown
		{
			get => _isOperatorAddressUnknown;
			set => Set(ref _isOperatorAddressUnknown, value);
		}

		[MaxLength(150)]
		public string OperatorAddressStreet
		{
			get => _operatorAddressStreet;
			set => Set(ref _operatorAddressStreet, value);
		}

		[MaxLength(50)]
		public string OperatorAddressCity
		{
			get => _operatorAddressCity;
			set => Set(ref _operatorAddressCity, value);
		}

		public State OperatorAddressState
		{
			get => _operatorAddressState;
			set => Set(ref _operatorAddressState, value);
		}

		[MaxLength(200)]
		public string OperatorAddressStateDescription
		{
			get => _operatorAddressStateDescription;
			set => Set(ref _operatorAddressStateDescription, value);
		}

		[MaxLength(20)]
		public string OperatorAddressPostalCode
		{
			get => _operatorAddressPostalCode;
			set => Set(ref _operatorAddressPostalCode, value);
		}

		public bool IsTrackOwnerNameUnknown
		{
			get => _isTrackOwnerNameUnknown;
			set => Set(ref _isTrackOwnerNameUnknown, value);
		}

		[MaxLength(50)]
		public string TrackOwnerName
		{
			get => _trackOwnerName;
			set => Set(ref _trackOwnerName, value);
		}

		public bool IsTrackOwnerAddressUnknown
		{
			get => _isTrackOwnerAddressUnknown;
			set => Set(ref _isTrackOwnerAddressUnknown, value);
		}

		[MaxLength(150)]
		public string TrackOwnerAddressStreet
		{
			get => _trackOwnerAddressStreet;
			set => Set(ref _trackOwnerAddressStreet, value);
		}

		[MaxLength(50)]
		public string TrackOwnerAddressCity
		{
			get => _trackOwnerAddressCity;
			set => Set(ref _trackOwnerAddressCity, value);
		}

		public State TrackOwnerAddressState
		{
			get => _trackOwnerAddressState;
			set => Set(ref _trackOwnerAddressState, value);
		}

		[MaxLength(200)]
		public string TrackOwnerAddressStateDescription
		{
			get => _trackOwnerAddressStateDescription;
			set => Set(ref _trackOwnerAddressStateDescription, value);
		}

		[MaxLength(20)]
		public string TrackOwnerAddressPostalCode
		{
			get => _trackOwnerAddressPostalCode;
			set => Set(ref _trackOwnerAddressPostalCode, value);
		}

		[ListCannotContainNull]
		public TrainWarningDeviceList WarningDevices { get; }

		[MaxLength(200)]
		public string WarningDevicesDescription
		{
			get => _warningDevicesDescription;
			set => Set(ref _warningDevicesDescription, value);
		}

		[ListCannotContainNull]
		public TrainAdvanceWarningDeviceList AdvanceWarningDevices { get; }

		[MaxLength(200)]
		public string AdvanceWarningDevicesDescription
		{
			get => _advanceWarningDevicesDescription;
			set => Set(ref _advanceWarningDevicesDescription, value);
		}

		[ListCannotContainNull]
		public TrainActiveWarningDeviceStatusList ActiveWarningDeviceStatuses { get; }

		[MaxLength(200)]
		public string ActiveWarningDeviceStatusesDescription
		{
			get => _activeWarningDeviceStatusesDescription;
			set => Set(ref _activeWarningDeviceStatusesDescription, value);
		}

		public TrainCrossingVehicleInteraction CrossingVehicleInteraction
		{
			get => _crossingVehicleInteraction;
			set => Set(ref _crossingVehicleInteraction, value);
		}

		public bool IsMakeUnknown
		{
			get => _isMakeUnknown;
			set => Set(ref _isMakeUnknown, value);
		}

		[MaxLength(25)]
		public string Make
		{
			get => _make;
			set => Set(ref _make, value);
		}

		public bool IsTypeUnknown
		{
			get => _isTypeUnknown;
			set => Set(ref _isTypeUnknown, value);
		}

		[MaxLength(25)]
		public string Type
		{
			get => _type;
			set => Set(ref _type, value);
		}

		public bool IsLeadEngineNumberUnknown
		{
			get => _isLeadEngineNumberUnknown;
			set => Set(ref _isLeadEngineNumberUnknown, value);
		}

		[MaxLength(20)]
		public string LeadEngineNumber
		{
			get => _leadEngineNumber;
			set => Set(ref _leadEngineNumber, value);
		}

		public bool IsSerialNumberUnknown
		{
			get => _isSerialNumberUnknown;
			set => Set(ref _isSerialNumberUnknown, value);
		}

		[MaxLength(25)]
		public string SerialNumber
		{
			get => _serialNumber;
			set => Set(ref _serialNumber, value);
		}

		public bool IsNumberOfEnginesUnknown
		{
			get => _isNumberOfEnginesUnknown;
			set => Set(ref _isNumberOfEnginesUnknown, value);
		}

		public int? NumberOfEngines
		{
			get => _numberOfEngines;
			set => Set(ref _numberOfEngines, value);
		}

		public bool IsNumberOfCarsUnknown
		{
			get => _isNumberOfCarsUnknown;
			set => Set(ref _isNumberOfCarsUnknown, value);
		}

		public int? NumberOfCars
		{
			get => _numberOfCars;
			set => Set(ref _numberOfCars, value);
		}

		public bool IsDistanceTraveledAfterImpactNotApplicable
		{
			get => _isDistanceTraveledAfterImpactNotApplicable;
			set => Set(ref _isDistanceTraveledAfterImpactNotApplicable, value);
		}

		public decimal? DistanceTraveledAfterImpact
		{
			get => _distanceTraveledAfterImpact;
			set => Set(ref _distanceTraveledAfterImpact, value);
		}

		public DistanceUnit? DistanceTraveledAfterImpactUnit
		{
			get => _distanceTraveledAfterImpactUnit;
			set => Set(ref _distanceTraveledAfterImpactUnit, value);
		}

		[ListCannotContainNull]
		public TrainEquipmentStatusList EquipmentStatuses { get; }

		public bool IsDataRecorderSpeedPending
		{
			get => _isDataRecorderSpeedPending;
			set => Set(ref _isDataRecorderSpeedPending, value);
		}

		public int? DataRecorderSpeed
		{
			get => _dataRecorderSpeed;
			set => Set(ref _dataRecorderSpeed, value);
		}

		public TrainCollisionType CollisionType
		{
			get => _collisionType;
			set => Set(ref _collisionType, value);
		}

		public bool IsStruckCarPositionNotApplicable
		{
			get => _isStruckCarPositionNotApplicable;
			set => Set(ref _isStruckCarPositionNotApplicable, value);
		}

		public bool IsStruckCarPositionUnknown
		{
			get => _isStruckCarPositionUnknown;
			set => Set(ref _isStruckCarPositionUnknown, value);
		}

		public int? StruckCarPosition
		{
			get => _struckCarPosition;
			set => Set(ref _struckCarPosition, value);
		}

		public bool IsStruckCarTypeNotApplicable
		{
			get => _isStruckCarTypeNotApplicable;
			set => Set(ref _isStruckCarTypeNotApplicable, value);
		}

		public bool IsStruckCarTypeUnknown
		{
			get => _isStruckCarTypeUnknown;
			set => Set(ref _isStruckCarTypeUnknown, value);
		}

		[MaxLength(25)]
		public string StruckCarType
		{
			get => _struckCarType;
			set => Set(ref _struckCarType, value);
		}

		public bool IsStruckCarNumberNotApplicable
		{
			get => _isStruckCarNumberNotApplicable;
			set => Set(ref _isStruckCarNumberNotApplicable, value);
		}

		public bool IsStruckCarNumberUnknown
		{
			get => _isStruckCarNumberUnknown;
			set => Set(ref _isStruckCarNumberUnknown, value);
		}

		[MaxLength(20)]
		public string StruckCarNumber
		{
			get => _struckCarNumber;
			set => Set(ref _struckCarNumber, value);
		}

		public HazardousMaterialsPlacardStatus HazardousMaterialsPlacardStatus
		{
			get => _hazardousMaterialsPlacardStatus;
			set => Set(ref _hazardousMaterialsPlacardStatus, value);
		}

		public bool IsHazardousMaterialsIdUnknown
		{
			get => _isHazardousMaterialsIdUnknown;
			set => Set(ref _isHazardousMaterialsIdUnknown, value);
		}

		public bool IsHazardousMaterialsIdNotApplicable
		{
			get => _isHazardousMaterialsIdNotApplicable;
			set => Set(ref _isHazardousMaterialsIdNotApplicable, value);
		}

		[MaxLength(20)]
		public string HazardousMaterialsId
		{
			get => _hazardousMaterialsId;
			set => Set(ref _hazardousMaterialsId, value);
		}

		public HazardousMaterialClass HazardousMaterialClass
		{
			get => _hazardousMaterialClass;
			set => Set(ref _hazardousMaterialClass, value);
		}

		public HazardousMaterialRelease HazardousMaterialRelease
		{
			get => _hazardousMaterialRelease;
			set => Set(ref _hazardousMaterialRelease, value);
		}

		[JsonSchemaComment("If the train had no engineer, set Engineer to null.")]
		public Engineer Engineer
		{
			get => _engineer;
			set => Set(ref _engineer, value);
		}

		[JsonSchemaComment("If the train had no conductor, set Conductor to null.")]
		public Conductor Conductor
		{
			get => _conductor;
			set => Set(ref _conductor, value);
		}

		[ListCannotContainNull]
		public TrainPassengerList Passengers { get; }

		void INeedsCollectionIndex.OnCollectionIndexChanged()
		{
			NotifyOfPropertyChange(nameof(Index));
			NotifyOfPropertyChange(nameof(Number));
		}
	}
}