﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;
using ECrash.Core.Models.Document;
using ECrash.Core.Models.ReferenceData;
using ECrash.Core.Models.Validation;
using ECrashLA.Console.Services.Rendering.Validation;
using ECrashLA.Console.Services.Serialization;
using ECrashLA.Console.Settings;
using ECrashLA.Core.Models.CrashReports;
using Newtonsoft.Json.Linq;
using Spectre.Console.Cli;

namespace ECrashLA.Console.Commands
{
	public class LocalValidateCrashReportCommand : AsyncCommand<ValidateLocalCrashReportSettings>
	{
		private readonly ICrashReportSerializer _serializer;
		private readonly IValidator<CrashReport> _validator;
		private readonly IValidationReportRenderer _renderer;
		private readonly IReferenceDataCache _referenceDataCache;
		private readonly IList<IBusinessRule<CrashReport>> _businessRules;

		public LocalValidateCrashReportCommand(
			ICrashReportSerializer serializer, 
			IValidator<CrashReport> validator, 
			IValidationReportRenderer renderer,
			IReferenceDataCache referenceDataCache,
			IList<IBusinessRule<CrashReport>> businessRules)
		{
			_serializer = serializer;
			_validator = validator;
			_renderer = renderer;
			_referenceDataCache = referenceDataCache;
			_businessRules = businessRules;
		}

		public override async Task<int> ExecuteAsync(CommandContext context, ValidateLocalCrashReportSettings settings)
		{
			var crashReport = await _serializer.LoadAsync(settings.CrashReportPath);

			var referenceDataJson = await File.ReadAllTextAsync(settings.RefereneDataPath);
			var token = JToken.Parse(referenceDataJson);
			ReferenceDataList referenceData;
			if (token is JObject obj)
			{
				if (obj.ContainsKey("ReferenceData"))
				{
					obj = (JObject)obj["ReferenceData"];
				}
				referenceData = obj.ToObject<ReferenceDataList>();
			}
			else
			{
				throw new Exception("No valid reference data JSON data found");
			}
			_referenceDataCache.Update(referenceData);
			
			FreshenCrashReport(crashReport);

			var validationContext = new Dictionary<string, object>();
			var validationReport = _validator.Validate(crashReport, validationContext);
			await _renderer.RenderAsync(validationReport, settings.Format);
			return validationReport.IsValid ? 0 : (int)validationReport.Status;
		}

		private void FreshenCrashReport(CrashReport crashReport)
		{
			using (crashReport.SuspendNotifications())
			{
				// Synchronize reference data values with the current ones on the server
				crashReport.SynchronizeReferenceData(_referenceDataCache.GetAll());

				// Apply business rules to recompute any computed fields
				crashReport.ApplyBusinessRulesOnce(_businessRules);
			}
		}

	}
}