﻿using System;
using System.Threading.Tasks;
using Castle.Core.Logging;
using ECrashLA.Console.Settings;
using Flurl;
using Flurl.Http;
using Microsoft.IdentityModel.JsonWebTokens;
using Spectre.Console;
using Spectre.Console.Cli;

namespace ECrashLA.Console.Commands
{
	public abstract class VendorCommandBase<T> : AsyncCommand<T> where T : AuthenticationSettings
	{
		protected VendorCommandBase(ILogger logger)
		{
			Logger = logger;
		}

		protected ILogger Logger { get; }
		
		public override async Task<int> ExecuteAsync(CommandContext context, T settings)
		{
			try
			{
				string token = await RenewTokenAsync(settings.Token, settings);
				return await ExecuteAsyncCore(context, settings, token);
			}
			catch (FlurlHttpException ex)
			{
				AnsiConsole.WriteException(ex);
				return ex.StatusCode ?? -1;
			}
			catch (Exception ex)
			{
				AnsiConsole.WriteException(ex);
				return -1;
			}
		}

		protected abstract Task<int> ExecuteAsyncCore(CommandContext context, T settings, string token);

		protected async Task<string> RenewTokenAsync(string token, T settings)
		{
			// Can't renew if we don't have the credentials
			if (settings.Username == null || settings.Password == null) return token;
			
			if (token != null)
			{
				var jwt = new JsonWebToken(token);
				if (jwt.ValidTo - DateTime.UtcNow > TimeSpan.FromMinutes(10))
				{
					// Token doesn't need to be renewed
					return token;
				}
				Logger.Debug("Renewing token...");
				token = await settings.Url
				                      .AppendPathSegment("vendor/token/v1")
				                      .PostJsonAsync(new { settings.Username, settings.Password })
				                      .ReceiveString();
			}
			return token;
		}
	}
}