﻿using System.IO;
using System.Threading.Tasks;
using Castle.Core.Logging;
using ECrashLA.Console.Models.CrashReports.Download;
using ECrashLA.Console.Settings;
using ECrashLA.Core.Models.CrashReports;
using Flurl;
using Flurl.Http;
using Newtonsoft.Json;
using Spectre.Console.Cli;

namespace ECrashLA.Console.Commands
{
	public class VendorDownloadCrashReportCommand : VendorCommandBase<VendorDownloadCrashReportSettings>
	{
		public VendorDownloadCrashReportCommand(ILogger logger) : base(logger)
		{
		}

		protected override async Task<int> ExecuteAsyncCore(CommandContext context, VendorDownloadCrashReportSettings settings, string token)
		{
			var serializer = JsonSerializer.CreateDefault();
			Directory.CreateDirectory(settings.OutputFolder);
			try
			{
				Logger.Debug($"Downloading crash report status and data for report {settings.Id}...");
				var result = await settings.Url.AppendPathSegments("vendor/download/v1/crashReport", settings.Id)
				                           .SetQueryParam("revision", settings.Revision)
										   .WithOAuthBearerToken(token)
										   .GetJsonAsync<DownloadCrashReportResult>();

				if (result.Status.ApprovalStatus != CrashReportApprovalStatus.Approved && !settings.DownloadUnapprovedReports)
				{
					Logger.Error($"Crash report {settings.Id} is not approved");
					return 451;
				}

				// Hook up the CrashReport and CrashReportStatus so things like the case number properties work
				result.CrashReport.Status = result.Status;

				string crashReportFolder = Path.Combine(settings.OutputFolder, settings.Id.ToString("D"));
				Directory.CreateDirectory(crashReportFolder);

				await using (var jsonFile = File.CreateText(Path.Combine(crashReportFolder, "CrashReport.json")))
				{
					serializer.Serialize(jsonFile, result.CrashReport);
				}
				await using (var jsonFile = File.CreateText(Path.Combine(crashReportFolder, "CrashReportStatus.json")))
				{
					serializer.Serialize(jsonFile, result.Status);
				}

				if (settings.DownloadDiagrams)
				{
					Logger.Debug($"Downloading diagrams for crash report {settings.Id}...");
					foreach (var diagram in result.CrashReport.Diagrams)
					{
						string diagramFolder = Path.Combine(crashReportFolder, "Diagrams", diagram.Hash.ToString());
						Directory.CreateDirectory(diagramFolder);
						Logger.Debug($"Downloading diagram {diagram.Filename} for crash report {settings.Id}...");
						await using var inputStream = await settings.Url.AppendPathSegments("vendor/download/v1/attachment", result.CrashReport.Id, diagram.Hash, diagram.Filename)
						                                            .WithOAuthBearerToken(token)
						                                            .GetStreamAsync();
						await using var outputStream = File.Create(Path.Combine(diagramFolder, diagram.Filename.ToString()));
						await inputStream.CopyToAsync(outputStream);
					}
				}

				if (settings.DownloadAttachments)
				{
					Logger.Debug($"Downloading attachments for crash report {settings.Id}...");
					foreach (var attachment in result.CrashReport.Attachments)
					{
						string attachmentFolder = Path.Combine(crashReportFolder, "Attachments", attachment.Hash.ToString());
						Directory.CreateDirectory(attachmentFolder);
						Logger.Debug($"Downloading attachment {attachment.Filename} for crash report {settings.Id}...");
						await using var inputStream = await settings.Url.AppendPathSegments("vendor/download/v1/attachment", result.CrashReport.Id, attachment.Hash, attachment.Filename)
						                                            .WithOAuthBearerToken(token)
						                                            .GetStreamAsync();
						await using var outputStream = File.Create(Path.Combine(attachmentFolder, attachment.Filename.ToString()));
						await inputStream.CopyToAsync(outputStream);
					}
				}

				if (settings.DownloadPdf)
				{
					Logger.Debug($"Downloading PDF for crash report {settings.Id}...");
					await using var inputStream = await settings.Url.AppendPathSegments("vendor/download/v1/pdf", settings.Id)
					                                            .SetQueryParam("revision", settings.Revision)
					                                            .SetQueryParam("includeAttachments", !settings.OmitAttachments)
					                                            .SetQueryParam("includePhotos", !settings.OmitPhotos)
					                                            .SetQueryParam("redact", settings.Redact)
					                                            .WithOAuthBearerToken(token)
					                                            .GetStreamAsync();
					await using var outputStream = File.Create(Path.Combine(crashReportFolder, "CrashReport.pdf"));
					await inputStream.CopyToAsync(outputStream);
				}
			}
			catch (FlurlHttpException ex) when (ex.StatusCode == 451)
			{
				Logger.Error($"Crash report {settings.Id} is not approved and you do not have permission to download it");
				return 451;
			}
			return 0;
		}
	}
}