﻿using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using Castle.Core.Logging;
using ECrashLA.Console.Settings;
using ECrashLA.Core.Models.CrashReports;
using Flurl;
using Flurl.Http;
using Humanizer;
using Newtonsoft.Json;
using Spectre.Console.Cli;

namespace ECrashLA.Console.Commands
{
	public class VendorDownloadCrashReportStatusesCommand : VendorCommandBase<VendorDownloadCrashReportStatusesSettings>
	{
		private static readonly Regex EtagRegex = new Regex(@"^(?:W/)?\""([0-9a-fA-F]{16})\""");

		public VendorDownloadCrashReportStatusesCommand(ILogger logger) : base(logger)
		{
		}

		protected override async Task<int> ExecuteAsyncCore(CommandContext context, VendorDownloadCrashReportStatusesSettings settings, string token)
		{
			string etag = settings.RowVersion ?? "0000000000000000";
			var allStatuses = new List<CrashReportStatus>();
			while (true)
			{
				Logger.Debug($"Downloading crash report statuses for {settings.AgencyId} newer than {etag}...");
				var response = await settings.Url.AppendPathSegments("vendor/download/v1/statuses", settings.AgencyId.ToString())
				                             .WithOAuthBearerToken(token)
				                             .WithHeader("If-None-Match", $"\"{etag}\"")
				                             .GetAsync();
				if (response.Headers.TryGetFirst("ETag", out string header))
				{
					var match = EtagRegex.Match(header);
					if (match.Success)
					{
						etag = match.Groups[1].Value;
					}
				}
				var statuses = await response.GetJsonAsync<List<CrashReportStatus>>();
				if (statuses == null || statuses.Count == 0) break;
				Logger.Debug($"Received {"crash report status".ToQuantity(statuses.Count)}; new RowVersion is {etag}.");
				allStatuses.AddRange(statuses);
			}
			using var jsonWriter = new JsonTextWriter(System.Console.Out) { CloseOutput = false };
			await jsonWriter.WriteStartObjectAsync();
			await jsonWriter.WritePropertyNameAsync("RowVersion");
			await jsonWriter.WriteValueAsync(etag);
			await jsonWriter.WritePropertyNameAsync("Statuses");
			var serializer = JsonSerializer.CreateDefault();
			serializer.Serialize(jsonWriter, allStatuses);
			await jsonWriter.WriteEndObjectAsync();

			return 0;
		}
	}
}