﻿using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using Castle.Core.Logging;
using ECrashLA.Console.Models.CrashReports.Download;
using ECrashLA.Console.Settings;
using ECrashLA.Core.Models.CrashReports;
using Flurl;
using Flurl.Http;
using Humanizer;
using Newtonsoft.Json;
using Spectre.Console.Cli;

namespace ECrashLA.Console.Commands
{
	public class VendorDownloadCrashReportsCommand : VendorCommandBase<VendorDownloadCrashReportsSettings>
	{
		private static readonly Regex EtagRegex = new(@"^(?:W/)?\""([0-9a-fA-F]{16})\""");

		public VendorDownloadCrashReportsCommand(ILogger logger) : base(logger)
		{
		}

		protected override async Task<int> ExecuteAsyncCore(CommandContext context, VendorDownloadCrashReportsSettings settings, string token)
		{
			var serializer = JsonSerializer.CreateDefault();
			string etag = settings.RowVersion ?? "0000000000000000";
			Directory.CreateDirectory(settings.OutputFolder);
			string rowVersionFile = Path.Combine(settings.OutputFolder, "RowVersion.txt");
			if (File.Exists(rowVersionFile) && (settings.RowVersion == null || settings.RowVersion == "0000000000000000"))
			{
				etag = await File.ReadAllTextAsync(rowVersionFile, Encoding.UTF8);
			}
			while (true)
			{
				token = await RenewTokenAsync(token, settings);
				Logger.Debug($"Downloading crash report statuses for {settings.AgencyId} newer than {etag}...");
				var response = await settings.Url.AppendPathSegments("vendor/download/v1/statuses", settings.AgencyId.ToString())
				                             .WithOAuthBearerToken(token)
				                             .WithHeader("If-None-Match", $"\"{etag}\"")
				                             .GetAsync();
				if (response.Headers.TryGetFirst("ETag", out string header))
				{
					var match = EtagRegex.Match(header);
					if (match.Success)
					{
						etag = match.Groups[1].Value;
					}
				}
				var statuses = await response.GetJsonAsync<List<CrashReportStatus>>();
				if (statuses == null || statuses.Count == 0) break;
				Logger.Debug($"Received {"crash report status".ToQuantity(statuses.Count)}; new RowVersion is {etag}.");

				foreach (var status in statuses)
				{
					token = await RenewTokenAsync(token, settings);
					if (status.ApprovalStatus != CrashReportApprovalStatus.Approved && !settings.DownloadUnapprovedReports)
					{
						Logger.Debug($"Crash report {status.Id} is not approved; skipping");
						continue;
					}

					try
					{
						Logger.Debug($"Downloading crash report status and data for report {status.Id}...");
						var result = await settings.Url.AppendPathSegments("vendor/download/v1/crashReport", status.Id)
						                           .WithOAuthBearerToken(token)
						                           .GetJsonAsync<DownloadCrashReportResult>();

						if (result.Status.ApprovalStatus != CrashReportApprovalStatus.Approved && !settings.DownloadUnapprovedReports)
						{
							Logger.Debug($"Crash report {status.Id} is not approved; skipping");
							continue;
						}

						// Hook up the CrashReport and CrashReportStatus so things like the case number properties work
						result.CrashReport.Status = result.Status;

						string crashReportFolder = Path.Combine(settings.OutputFolder, status.Id.ToString("D"));
						Directory.CreateDirectory(crashReportFolder);

						await using (var jsonFile = File.CreateText(Path.Combine(crashReportFolder, "CrashReport.json")))
						{
							serializer.Serialize(jsonFile, result.CrashReport);
						}
						await using (var jsonFile = File.CreateText(Path.Combine(crashReportFolder, "CrashReportStatus.json")))
						{
							serializer.Serialize(jsonFile, result.Status);
						}

						if (settings.DownloadDiagrams && result.CrashReport.Diagrams.Count > 0)
						{
							Logger.Debug($"Downloading diagrams for crash report {status.Id}...");
							foreach (var diagram in result.CrashReport.Diagrams)
							{
								string diagramFolder = Path.Combine(crashReportFolder, "Diagrams", diagram.Hash.ToString());
								Directory.CreateDirectory(diagramFolder);
								Logger.Debug($"Downloading diagram {diagram.Filename} for crash report {status.Id}...");
								await using var inputStream = await settings.Url.AppendPathSegments("vendor/download/v1/attachment", result.CrashReport.Id, diagram.Hash, diagram.Filename)
								                                            .WithOAuthBearerToken(token)
								                                            .GetStreamAsync();
								await using var outputStream = File.Create(Path.Combine(diagramFolder, diagram.Filename.ToString()));
								await inputStream.CopyToAsync(outputStream);
							}
						}

						if (settings.DownloadAttachments && result.CrashReport.Attachments.Count > 0)
						{
							Logger.Debug($"Downloading attachments for crash report {status.Id}...");
							foreach (var attachment in result.CrashReport.Attachments)
							{
								string attachmentFolder = Path.Combine(crashReportFolder, "Attachments", attachment.Hash.ToString());
								Directory.CreateDirectory(attachmentFolder);
								Logger.Debug($"Downloading attachment {attachment.Filename} for crash report {status.Id}...");
								await using var inputStream = await settings.Url.AppendPathSegments("vendor/download/v1/attachment", result.CrashReport.Id, attachment.Hash, attachment.Filename)
								                                            .WithOAuthBearerToken(token)
								                                            .GetStreamAsync();
								await using var outputStream = File.Create(Path.Combine(attachmentFolder, attachment.Filename.ToString()));
								await inputStream.CopyToAsync(outputStream);
							}
						}

						if (settings.DownloadPdf)
						{
							Logger.Debug($"Downloading PDF for crash report {status.Id}...");
							await using var inputStream = await settings.Url.AppendPathSegments("vendor/download/v1/pdf", status.Id)
							                                            .SetQueryParam("revision", 0)
							                                            .SetQueryParam("includeAttachments", !settings.OmitAttachments)
							                                            .SetQueryParam("includePhotos", !settings.OmitPhotos)
							                                            .SetQueryParam("redact", settings.Redact)
							                                            .WithOAuthBearerToken(token)
							                                            .GetStreamAsync();
							await using var outputStream = File.Create(Path.Combine(crashReportFolder, "CrashReport.pdf"));
							await inputStream.CopyToAsync(outputStream);
						}
					}
					catch (FlurlHttpException ex) when (ex.StatusCode == 451)
					{
						Logger.Debug($"Crash report {status.Id} is not approved and you do not have permission to download it; skipping");
					}
				}
				await File.WriteAllTextAsync(rowVersionFile, etag, Encoding.UTF8);
			}
			return 0;
		}
	}
}