﻿using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using Castle.Core.Logging;
using ECrashLA.Console.Settings;
using Flurl;
using Flurl.Http;
using Humanizer;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using Spectre.Console.Cli;

namespace ECrashLA.Console.Commands;

public class VendorDownloadFilteredCrashReportDataCommand : VendorCommandBase<VendorDownloadFilteredCrashReportDataSettings>
{
	private static readonly Regex EtagRegex = new Regex(@"^(?:W/)?\""([0-9a-fA-F]{16})\""");

	public VendorDownloadFilteredCrashReportDataCommand(ILogger logger) : base(logger)
	{
	}

	protected override async Task<int> ExecuteAsyncCore(CommandContext context, VendorDownloadFilteredCrashReportDataSettings settings, string token)
	{
		string etag = settings.RowVersion ?? "0000000000000000";
		var allStatuses = new List<JToken>();
		while (true)
		{
			Logger.Debug($"Downloading filtered crash report data for {settings.AgencyId} using filter {settings.Filter} newer than {etag}...");
			var response = await settings.Url.AppendPathSegments("vendor/download/v1/filteredReports", settings.AgencyId.ToString(), settings.Filter)
			                             .WithOAuthBearerToken(token)
			                             .WithHeader("If-None-Match", $"\"{etag}\"")
			                             .GetAsync();
			if (response.Headers.TryGetFirst("ETag", out string header))
			{
				var match = EtagRegex.Match(header);
				if (match.Success)
				{
					etag = match.Groups[1].Value;
				}
			}
			var reports = await response.GetJsonAsync<JArray>();
			if (reports == null || reports.Count == 0) break;
			Logger.Debug($"Received {"filtered crash reports".ToQuantity(reports.Count)}; new RowVersion is {etag}.");
			allStatuses.AddRange(reports);
		}
		using var jsonWriter = new JsonTextWriter(System.Console.Out) { CloseOutput = false };
		await jsonWriter.WriteStartObjectAsync();
		await jsonWriter.WritePropertyNameAsync("RowVersion");
		await jsonWriter.WriteValueAsync(etag);
		await jsonWriter.WritePropertyNameAsync("CrashReports");
		var serializer = JsonSerializer.CreateDefault();
		serializer.Serialize(jsonWriter, allStatuses);
		await jsonWriter.WriteEndObjectAsync();

		return 0;
	}
}