﻿using System;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ECrash.Core;
using ECrash.Core.Ioc;
using ECrash.Core.Models.Document;
using ECrashLA.Core.Models.CrashReports;
using Newtonsoft.Json;

namespace ECrashLA.Console.Services.Serialization
{
	[Singleton]
	public class CrashReportSerializer : ICrashReportSerializer
	{
		public Task<CrashReport> LoadAsync(string path)
		{
			if (Directory.Exists(path))
			{
				return LoadFromDirectoryAsync(path);
			}
			if (!File.Exists(path))
			{
				throw new FileNotFoundException("No file or directory containing crash report data found", path);
			}
			if (Path.GetExtension(path).Equals(".json", StringComparison.OrdinalIgnoreCase))
			{
				string directory = Path.GetDirectoryName(path);
				string filename = Path.GetFileName(path);
				return LoadFromDirectoryAsync(directory, filename);
			}
			return LoadFromFileAsync(path);
		}

		private async Task<CrashReport> LoadFromDirectoryAsync(string path, string jsonFilename = "CrashReport.json")
		{
			string jsonPath = Path.Combine(path, jsonFilename);

			using var reader = File.OpenText(jsonPath);
			using var jsonReader = new JsonTextReader(reader);
			var serializer = JsonSerializer.CreateDefault();
			var crashReport = serializer.Deserialize<CrashReport>(jsonReader);
			if (crashReport == null) throw new InvalidOperationException("Unable to deserialize CrashReport JSON");
			foreach (var attachment in crashReport.EnumerateAttachments())
			{
				foreach (var file in Directory.EnumerateFiles(path, attachment.Filename.ToString(), SearchOption.AllDirectories))
				{
					var data = await File.ReadAllBytesAsync(file);
					var hash = Hash.ComputeHash(data);
					if (hash == attachment.Hash)
					{
						attachment.Data = data;
						break;
					}
				}
				if (attachment.Data == null)
				{
					throw new FileNotFoundException($"Missing attachment file \"{attachment.Filename}\" with hash {attachment.Hash}");
				}
			}
			return crashReport;
		}

		private async Task<CrashReport> LoadFromFileAsync(string path)
		{
			using var zipfile = ZipFile.OpenRead(path);
			var jsonEntry = zipfile.GetEntry("CrashReport.json") ?? throw new InvalidOperationException($"Missing CrashReport.json file in {path}");
			await using var jsonStream = jsonEntry.Open();
			using var reader = new StreamReader(jsonStream, Encoding.UTF8);
			using var jsonReader = new JsonTextReader(reader);
			var serializer = JsonSerializer.CreateDefault();
			var crashReport = serializer.Deserialize<CrashReport>(jsonReader);
			if (crashReport == null) throw new InvalidOperationException("Unable to deserialize CrashReport JSON");
			foreach (var attachment in crashReport.EnumerateAttachments())
			{
				foreach (var entry in zipfile.Entries.Where(x => x.Name == (string)attachment.Filename))
				{
					await using var entryStream = entry.Open();
					var data = await entryStream.ReadToEndAsync();
					var hash = Hash.ComputeHash(data);
					if (hash == attachment.Hash)
					{
						attachment.Data = data;
						break;
					}
				}
				if (attachment.Data == null)
				{
					throw new FileNotFoundException($"Missing attachment file \"{attachment.Filename}\" with hash {attachment.Hash}");
				}
			}
			return crashReport;
		}
	}
}